
import re 


def postprocess_sparql_tokens(sparql_query):
    sparql_query = re.sub(r'(\?[a-zA-Z0-9_]+)', r' \1 ', sparql_query)
    # 去除多余的空格
    sparql_query = re.sub(r'\s+', ' ', sparql_query)
    return sparql_query


def detect_query_language(query_string):
    # 判断是否包含SPARQL关键词
    sparql_keywords = ['SELECT', 'DISTINCT', 'WHERE', 'FILTER', 'OPTIONAL', 'ORDER BY', 'LIMIT', 'ASK']
    has_sparql_keywords = any(keyword in query_string.upper() for keyword in sparql_keywords)

    # 判断是否包含Cypher关键词
    cypher_keywords = ['MATCH', 'OPTIONAL MATCH', 'WHERE', 'RETURN', 'ORDER BY', 'LIMIT']
    has_cypher_keywords = any(keyword in query_string.upper() for keyword in cypher_keywords)

    if has_sparql_keywords and not has_cypher_keywords:
        return 'sparql'
    elif has_cypher_keywords and not has_sparql_keywords:
        return 'cypher'
    else:
        raise Exception(f"unknown query language `{query_string}`")




# -*- coding: utf-8 -*-
# @File    :   utils.py
# @Time    :   2023/03/13 14:40:16
# @Author  :   Qing 
# @Email   :   aqsz2526@outlook.com
######################### docstring ########################
'''

'''

from antlr4.tree.Tree import TerminalNodeImpl
from .cypher.utils.parse import parse_cypher
from .sparql.utils.parse import parse_sparql
from .utils import parse_layer

_CY_NODES = [
    # 'oC_SymbolicName',              # pv, e_1
    'oC_RelationshipTypes',         # :for_work :instance_of
    # 'oC_NodeLabel',                 # : Resource :Relation
    # 'oC_RelationshipDetail',            # [:instance_of]
    # 'oC_PropertyKeyName',           # value
    # 'oC_NumberLiteral',
    # 'oC_MapLiteral',                   # {name:"xxx"}
    # 'oC_Expression', 
    'oC_Literal'                    # "a string or number"
]

_SP_NODES = [
    # 'var',                          # ?v ?pv
    # 'pathPrimary',                  # <pred:value> date name fact_r h t 
    'iri',                          #  <for_work>
    'string',                       # "a string"
    'numericLiteralUnsigned',       # 1895
    'numericLiteralNegative'        # -192
]

def parse_subtree(tree, parts):
    if tree.getText() == "<EOF>":
        return
    elif isinstance(tree, TerminalNodeImpl) or tree.children is None:
        parts.append(tree.getText())
    else:
        for child in tree.children:
            parse_subtree(child, parts)

def get_subtree_leaves(tree):
    """ 遍历一个子树、返回其叶子结点的list """
    parts = [] 
    parse_subtree(tree, parts)
    return parts

# classnames = load_json("/home/qing/workspace/papers/query_skeleton/src/data_utils/class_names.json")
classnames = []

def parse_skeleton(tree, parts, ruleNames, idx2value, lang='sparql'):
    """
        将 ast 解析为骨架，保留 AST2Skeleton._SP_NODES 中定义的非终结符
        parts: 骨架节点
        idx2value : 非终结符到骨架节点idx 到 其叶子结点的映射
    """
    if tree.getText() == "<EOF>":
        return
    elif isinstance(tree, TerminalNodeImpl):
        parts.append(tree.getText())
    elif tree.children is None:
        return 
    else:
        nonterm_name = ruleNames[tree.getRuleIndex()]
        # print(lang, nonterm_name, AST2Skeleton.nonterms(lang))
        if nonterm_name in _CY_NODES + _SP_NODES:  # 如果是预定义的 非终结符
            values = get_subtree_leaves(tree)
            # print(len(parts)-1, nonterm_name, values)
            # if values[0][1:-1] in ["human", "percentage", "film", "city", "square metre", "square kilometre", "county", "metre"]:
            if values[0][1:-1] in classnames:
                pass  # instance of or unit 不算需要替换的
            else:
                parts.append(nonterm_name)
                idx2value[len(parts)-1] = (nonterm_name, values)
        else:
            for child in tree.children:
                parse_skeleton(child, parts, ruleNames, idx2value, lang=lang)


def parse_nodes_relations_cql(cy):
    """
    从Cypher中解析出节点和关系
    nodes:      ['{name:"human"}', '{value:"TheOliverStone"}', '{name:"Nixon"}', '{name:"Twitter_username"}']
    relations:  ['[:instance_of]', '[:Twitter_username]', '[:director]', '[:fact_h]', '[:fact_t]', '[:fact_r]', '[:number_of_subscribers]']
    """
    tree, parser = parse_cypher(cy,True)
    parts, d = [], {} 
    parse_skeleton(tree, parts, parser.ruleNames, d, lang='cypher')
    # print(parts)
    # print(d)
    nodes, relations = [], []
    for k,v in d.items():
        nonterm_name, values = v 
        if nonterm_name == "oC_RelationshipDetail" or nonterm_name == "oC_RelationshipTypes":
            relations.append( "".join(values) )
        # elif nonterm_name == "oC_MapLiteral":
        # elif nonterm_name == "oC_Expression":
            # nodes.append( "".join(values) )
        elif nonterm_name == "oC_Literal":
            nodes.append( "".join(values) )
        else:
            raise Exception(f'{nonterm_name} not handled!')
    
    # print(nodes)
    # print(relations)
    return nodes, relations 

def parse_nodes_relations_sparql(sp):
    """ 
        从sparql中解析出节点和关系
        nodes:      ['"human"', '"TheOliverStone"', '"Nixon"', '"Twitter_username"']
        relations:  ['<instance_of>', '<Twitter_username>', '<director>', '<fact_h>', '<fact_t>', '<fact_r>', '<number_of_subscribers>']
    """
    tree, parser = parse_sparql(sp,True)
    parts, d = [], {} 
    parse_skeleton(tree, parts, parser.ruleNames, d)
    # print(parts)
    # print(d)
    nodes, relations = [], []
    for k,v in d.items():
        nonterm_name, values = v 
        if nonterm_name in ["numericLiteralUnsigned", "numericLiteralNegative", "string"]:
            nodes.append( "".join(values) )
        elif nonterm_name in ['iri']:
            relations.append( "".join(values) )
        else:
            raise Exception(f'{nonterm_name} not handled!')
    
    # print(nodes)
    # print(relations)
    return nodes, relations 


def tokenize_sequence(tokenizer, list_of_str):
    """ 
        bartTokenizer for multi str
    """
    input_ids = [tokenizer.bos_token_id]
    for s in list_of_str:
        ids = tokenizer.encode(s)
        input_ids.extend(ids[1:]+ [tokenizer.eos_token_id])
        
    input_ids = input_ids[:-1]
    attention_mask = [1 for _ in input_ids]

    return input_ids, attention_mask



def tokenize_with_type_ids(tokenizer, input_ids, gql, offsets, lang='cypher'):
    """ 
        0 for normal, 1 for node , 2 for relation
    """
    node_target = [0 for _ in input_ids]
    rel_target = [0 for _ in input_ids]

    if lang == 'cypher':
        nodes, relations = parse_nodes_relations_cql(gql)
    elif lang == 'sparql':
        nodes, relations = parse_nodes_relations_sparql(gql)
    else:   
        raise Exception(f"{lang} not supported!")

    # for x in nodes:
    #     flag = False
    #     s = tokenizer(x).input_ids[1:-1]
    #     l = len(s)
    #     for i in range(1, len(input_ids)-l):
    #         if input_ids[i:i+l] == s:
    #             rel_target[i:i+l] = [1 for _ in s]
    #             flag = True
    #     if flag == False:
    #         raise Exception(f"No matched: {x} in {gql}")
    
    for n in nodes:
        s = gql.index(n)
        e = s+len(n)
        # print(s,e, gql[s:e])
        ts, te = None, None
        for i,(_s,_e) in enumerate(offsets):
            if _s == s or (_s < s <= _e):
                ts = i
            if _e == e or (_s <= e < _e):
                te = i 
        if ts is None or te is None:
            raise Exception(f"No matched: {n} in {gql}\n{(s,e)} | {(ts, te)} not in {offsets}")
        else:
            node_target[ts:te+1] = [1 for _ in range(ts,te+1)]

    for n in relations:
        s = gql.index(n)
        e = s+len(n)
        # print(s,e, gql[s:e])
        ts, te = None, None
        for i,(_s,_e) in enumerate(offsets):
            if _s == s or (_s < s <= _e):
                ts = i
            if _e == e or (_s <= e < _e):
                te = i 
        if ts is None or te is None:
            raise Exception(f"No matched: {n} in {gql}\n{(s,e)} | {(ts, te)} not in {offsets}")
        else:
            rel_target[ts:te+1] = [1 for _ in range(ts,te+1)]
            

    merged = [ 1 if n or r else 0 for n,r in zip(node_target, rel_target) ]

    return node_target, rel_target, merged




from ..register import registers

from qdls.gql.process_utils import parse_nodes_relations_sparql

@registers.process_function.register("kqa_test_function")
def kqa_test_function():
    """Test function."""
    print(f"This is a test function in kqa .")
    return True

@registers.process_function.register
def tokenization_for_SFT(example, tokenizer, mode='train'):
    """ 处理key为prompt, response的数据集 """
    if mode == 'test':
        input_str = example['prompt']
    else:
        input_str = example['prompt'] + example['response']

    td = tokenizer(input_str, truncation=True)
    example.update(td)
    if mode != 'test':
        example['labels'] = td.input_ids
    return example

@registers.process_function.register 
def kqa_cql_tokenization_causal(example, tokenizer, mode):
    """
        对一条数据的处理: 字符串处理、分词
    """
    if mode != 'test':
        text = "Here is a question: " + example['question'] \
            + " The CQL for this question is: " + example['cypher'] + tokenizer.eos_token
        td = tokenizer(text,  padding=False)
    else:
        text = "Here is a question: " + example['question'] + " The CQL for this question is: "
        td = tokenizer(text,  padding=False)
 
    d = {
        'input_ids': td.input_ids, 'attention_mask': td.attention_mask,
        'labels' :  td.input_ids if mode != 'test' else None,
    }
    d = dict(**example, **d)
    return d  


@registers.process_function.register 
def kqa_cql_tokenization_causal_with_gold_entity(example, tokenizer, mode):
    """ 
        golden entity extract from sparql
    """
    nodes, rels = parse_nodes_relations_sparql(example['sparql'])
    gold_ent = ", ".join(nodes)
    input_str = f" Related entities are {gold_ent} ."

    if mode != 'test':
        text = "Here is a question: " + example['question'] + input_str \
            + " The CQL for this question is: " + example['cypher'] + tokenizer.eos_token
        td = tokenizer(text,  padding=False)
    else:
        text = "Here is a question: " + example['question'] + input_str + " The CQL for this question is: "
        td = tokenizer(text,  padding=False)

    
    # td = tokenizer(input_str, padding=False)
    d = {
        'input_ids': td.input_ids, 'attention_mask': td.attention_mask,
        'labels' : None if mode == 'test' else td.input_ids
    }
    d = dict(**example, **d)
    return d 


@registers.process_function.register
def kqa_cql_tokenization_instruction_causal_with_gold_entity(example, tokenizer, mode):
    """ 
        以对话指令的形式，生成Cypher
    """
 
    nodes, rels = parse_nodes_relations_sparql(example['sparql'])
    gold_ent = ", ".join(nodes)

    if mode != 'test':
        text = f"User: here is a question: {example['question']}. Related entities are {gold_ent}. Give me the Cypher query for this question.\nAssistant:" + \
                f"{example['cypher']}" + tokenizer.eos_token
        td = tokenizer(text,  padding=False)
    else:
        text = f"User: here is a question: {example['question']}. Related entities are {gold_ent}. Give me the Cypher query for this question.\nAssistant:"
        td = tokenizer(text,  padding=False)

    d = {
        'input_ids': td.input_ids, 'attention_mask': td.attention_mask,
        'labels' : None if mode == 'test' else td.input_ids
    }
    d = dict(**example, **d)
    return d 


@registers.process_function.register
def kqa_cql_two_turns_with_gold_entitiy_and_feedback(example, tokenizer, mode):
    """ 
        User: here is a question: $Q. Related entities are $E. Give me the Cypher query for this question.
        Assistant: $Gen_CQL
        User: The DBMS returns $FeedBack. Try again.
        Assistant: $Gold_CQL
    
    """
    nodes, rels = parse_nodes_relations_sparql(example['sparql'])
    gold_ent = ", ".join(nodes)
    # input_str = f" Related entities are {gold_ent} ."

    def post_process_query(text):
        """ post process gpt generated query 
            pad pad Question The CQL for this question is: query => query
        """
        prefix = "The CQL for this question is:" 
        idx = text.rindex(prefix)
        query = text[idx + len(prefix):]
        return query.strip()


    if mode != 'test':
        rejected = post_process_query(example['rejected'])
        context = f"User: here is a question: {example['question']}. Related entities are {gold_ent}. Give me the Cypher query for this question.\n" + \
                f"Assistant: {rejected}\nUser: The DBMS returns {example['feedback']}. Try again.\n"
        text = context + f"Assistant: {example['cypher']}" + tokenizer.eos_token
        td = tokenizer(text,  padding=False)
    else:
        text = f"User: here is a question: {example['question']}. Related entities are {gold_ent}. Give me the Cypher query for this question.\nAssistant:"
        td = tokenizer(text,  padding=False)

    
    # td = tokenizer(input_str, padding=False)
    d = {
        'input_ids': td.input_ids, 'attention_mask': td.attention_mask,
        'labels' : None if mode == 'test' else td.input_ids
    }
    d = dict(**example, **d)
    return d 

@registers.process_function.register
def kqa_sparql_tokenization_instruction_causal_with_gold_entity(example, tokenizer, mode):
    """ 
        以对话指令的形式，生成sparql
    """
 
    nodes, rels = parse_nodes_relations_sparql(example['sparql'])
    gold_ent = ", ".join(nodes)
    input_str = f" Related entities are {gold_ent} ."

    if mode != 'test':
        text = f"User: here is a question: {example['question']}. Related entities are {gold_ent}. Give me the SPARQL query for this question.\nAssistant:" + \
                f"{example['sparql']}" + tokenizer.eos_token
        td = tokenizer(text,  padding=False)
    else:
        text = f"User: here is a question: {example['question']}. Related entities are {gold_ent}. Give me the SPARQL query for this question.\nAssistant:"
        td = tokenizer(text,  padding=False)

    d = {
        'input_ids': td.input_ids, 'attention_mask': td.attention_mask,
        'labels' : None if mode == 'test' else td.input_ids
    }
    d = dict(**example, **d)
    return d 


@registers.process_function.register
def kqa_sparql_two_turns_with_gold_entitiy_and_feedback(example, tokenizer, mode):
    """ 
        User: here is a question: $Q. Related entities are $E. Give me the SPARQL query for this question.
        Assistant: $Gen_SPARQL
        User: The DBMS returns $FeedBack. Try again.
        Assistant: $Gold_SPARQL
    
    """
    nodes, rels = parse_nodes_relations_sparql(example['sparql'])
    gold_ent = ", ".join(nodes)
    # input_str = f" Related entities are {gold_ent} ."

    def post_process_query(text):
        """ post process gpt generated query 
            pad pad Question The SPARQL for this question is: query => query
        """
        prefix = "The SPARQL for this question is:" 
        idx = text.rindex(prefix)
        query = text[idx + len(prefix):]
        return query.strip()


    if mode != 'test':
        rejected = post_process_query(example['rejected'])
        context = f"User: here is a question: {example['question']}. Related entities are {gold_ent}. Give me the SPARQL query for this question.\n" + \
                f"Assistant: {rejected}\nUser: The DBMS returns {example['feedback']}. Try again.\n"
        text = context + f"Assistant: {example['sparql']}" + tokenizer.eos_token
        td = tokenizer(text,  padding=False)
    else:
        text = f"User: here is a question: {example['question']}. Related entities are {gold_ent}. Give me the SPARQL query for this question.\nAssistant:"
        td = tokenizer(text,  padding=False)

    d = {
        'input_ids': td.input_ids, 'attention_mask': td.attention_mask,
        'labels' : None if mode == 'test' else td.input_ids
    }
    d = dict(**example, **d)
    return d 






import os 
import sys 

import datasets 
import torch
from torch.utils.data import DataLoader
from transformers import AutoTokenizer 
import pytorch_lightning as pl
from typing import * 

from ...utils import print_string, print_config
from .dataset_builder import DatasetBuilder
from ..register import registers


@registers.datamodule.register("base_dm")
class BaseDataModule(pl.LightningDataModule):
    """ 
        config:
            pretrained
            data:
                train_bsz
                val_bsz
                test_bsz
                padding_side
                collator_name
                tokenize_fn_name
                cache_dir
                force_reload
        需要先在registers中注册collator和process_function
        DataBuilder 需要 train_path val_path test_path 或者 dataset_name
    """
    def __init__(self, config, **kwargs) -> None:
        super().__init__() 

        self.config = config 
        print_string("configuration of datamodule")
        print_config(self.config.data)

        self.tokenizer = AutoTokenizer.from_pretrained(
            config.pretrained,
            trust_remote_code=True, 
            use_fast=getattr(config.data, 'use_fast', False),
            padding_side = config.data.padding_side
        )
        
        # 有些模型的tokenizer没有设置 pad token
        if self.tokenizer.pad_token_id is None:
            self.tokenizer.pad_token_id = self.tokenizer.eos_token_id 


        self.collator_cls = registers.collator.get(self.config.data.collator_name) 
        self.tokenize_fn = registers.process_function.get(self.config.data.tokenize_fn_name)

        self.num_workers = kwargs.get('num_workers', 4)

    def prepare_data(self, mode='train') -> None:
        """设置缓存文件的路径，如果缓存文件不存在则调用DatasetBuilder进行构建

        Args:
            mode: train or test. Defaults to 'train'.

        """
        self.cached_train = os.path.join(self.config.data.cache_dir, f"train_{self.config.data.tokenize_fn_name}")
        self.cached_val = os.path.join(self.config.data.cache_dir, f"val_{self.config.data.tokenize_fn_name}")
        self.cached_test = os.path.join(self.config.data.cache_dir, f"test_{self.config.data.tokenize_fn_name}")

        if not os.path.exists(self.config.data.cache_dir):
            os.makedirs(self.config.data.cache_dir, exist_ok=True)

        # 强制 or 训练时没有self.cached_train or 预测时没有self.cached_test
        if self.config.data.force_reload or ( (not os.path.exists(self.cached_train)) and mode=='train') \
            or ( (not os.path.exists(self.cached_test)) and mode =='test'):
            ds = DatasetBuilder(self.config)
            if mode == 'train':
                trainset = ds.build('train', self.tokenizer, self.tokenize_fn)
                valset =  ds.build('val', self.tokenizer, self.tokenize_fn)
                
                if trainset is not None:
                    trainset.save_to_disk(self.cached_train)
                if valset is not None:
                    valset.save_to_disk(self.cached_val)
            elif mode == 'test':
            
                testset = ds.build('test', self.tokenizer, self.tokenize_fn)
                assert testset is not None 
                testset.save_to_disk(self.cached_test)
            else:
                raise Exception(f'mode {mode} not imple')
            
            print_string("Data cache re-generated!")

    def setup(self, stage: Optional[str] = None) -> None:
        
        if stage == 'fit' or stage is None:
            self.trainset = datasets.load_from_disk(self.cached_train)
            self.valset = datasets.load_from_disk(self.cached_val)

        if stage == 'test' or stage is None:
            self.testset = datasets.load_from_disk(self.cached_test)
        
        print_string("Datasets setup finished!")

    def train_dataloader(self):
        return DataLoader(self.trainset, batch_size=self.config.data.train_bsz,
            shuffle=True,
            collate_fn=self.collator_cls(self.tokenizer, mode='train'),
            num_workers=self.num_workers,
        )

    def val_dataloader(self):
        return DataLoader(self.valset, batch_size=self.config.data.val_bsz, 
            shuffle=False,
            collate_fn=self.collator_cls(self.tokenizer, mode='val'),
            num_workers=self.num_workers,
        )

    def test_dataloader(self):
        return DataLoader(self.testset, batch_size=self.config.data.test_bsz, 
            shuffle=False,
            collate_fn=self.collator_cls(self.tokenizer, mode='test'),
            num_workers=self.num_workers
        )





# -*- coding: utf-8 -*-
# @File    :   base.py
# @Time    :   2023/11/08 13:50:38
# @Author  :   Qing 
# @Email   :   aqsz2526@outlook.com
######################### docstring ########################
'''
'''


####################### imports ###########################

from abc import ABC, abstractmethod
import datasets 
from tqdm import tqdm 
from collections import defaultdict
from qdls.data import load_json, save_json
from qdls.utils import print_string
from qdls.kgqa_eval.evaluator.metric_fns import calc_metrics_per_sample

class BaseEvaluator(ABC):
    """ 
        实现 __init__ 和 normalize_sample 函数
    """
    def __init__(self, file=None, refer_key='cypher', nproc=1, **kwargs) -> None:
        self.refer_key = refer_key
        self.nproc = nproc
        self.metrics_to_calc = []
        self._manage_metrics_to_calc(kwargs)
        print_string(f"metrics to calc: {self.metrics_to_calc}")
        print_string(f"neo4j config: {kwargs.get('neo4j_config', '|| please set neo4j_config in kwargs ||')}")

        if file is not None:
            self.raw_data = load_json(file) if type(file) is str else file 
            
            self.data_to_eval = [ self.normalize_sample(s, refer_key) for s in self.raw_data ]
            self._pre_calc_metrics()
            self.id2sample = {sample['sample_id']:sample for sample in self.data_to_eval}

            

    @abstractmethod
    def normalize_sample(self, sample, *args, **kwargs):
        """ standardize the sample to a dict with keys 
            sample_id

        IMPORTANT: values should not be dict, or it will be extremely slow
        """
        return sample
    
    def _manage_metrics_to_calc(self, kwargs):
        """  
            增加在 init 函数通过 kwargs 传入新的 metrics_to_calc
            lang 是 cypher 和 sparql 默认增加 is_correct
        """
        # if self.refer_key in ['cypher', 'sparql']:
        #     self.metrics_to_calc.append("is_correct")  # 默认是要计算准确率的 TODO:但是对应的是 kqa 的计算，考虑删除
        if "metrics_to_calc" in kwargs:
            for metric in kwargs['metrics_to_calc']:
                if metric not in self.metrics_to_calc:
                    self.metrics_to_calc.append(metric)   

    def evaluate(self, verbose=True):
        """ 读取已经计算完毕的 metrics 数据，在数据集上计算 macro 指标，并打印出来"""
        scores = self.calc_marco_metrics()
        if verbose:
            for k,v in scores.items():
                print(f"{k:<20}: {v:.4f}")
        return scores
    
    def _pre_calc_metrics(self):
        """ 使用 datasets库 并行计算所有指标 """
        ds = datasets.Dataset.from_list(self.data_to_eval)
        print(ds)
        
        if self.nproc > 1:
            ds = ds.map(
                calc_metrics_per_sample, num_proc=self.nproc, 
                fn_kwargs={
                    'ref_key':self.refer_key, 
                    'metrics':self.metrics_to_calc,
                    'neo4j_config': self.neo4j_config
                }
            )
            self.data_to_eval = ds.to_list()
        else:
            self.data_to_eval = [] 
            for s in tqdm(ds):
                self.data_to_eval.append(
                    calc_metrics_per_sample(s, self.refer_key, self.metrics_to_calc, self.neo4j_config)
                )
    
    def calc_marco_metrics(self, ids=None):
        """ 根据预先计算好的metrics, 为ids中的样本计算平均值 """
        if ids is None:
            ids = self.id2sample.keys()
        L = len(ids)
        scores = defaultdict(list)
        for i in ids:
            sample = self.id2sample[i]    
            for m in self.metrics_to_calc:
                scores[m].append(float(sample.get(m, 0)))
        return {k:sum(v)/L for k,v in scores.items()}
    
    def save_with_results(self, target_path):
        save_json(self.data_to_eval, target_path)  
    
 


# -*- coding: utf-8 -*-
# @File    :   pretrains.py
# @Time    :   2021/11/02 19:54:48
# @Author  :   Qing 
# @Email   :   sqzhao@stu.ecnu.edu.cn
"""
    本模块使用方法：
        执行saved_pretrains() 查看本地已经保存的模型
        想要的模型没有保存，可以通过models[index].save()保存到本地
"""
import os
import time
from dataclasses import dataclass
from transformers import AutoModel, AutoTokenizer, PreTrainedModel, AutoConfig
from transformers import AutoModelWithLMHead, AutoModelForSeq2SeqLM
from transformers import AutoModelForCausalLM, AutoModelForMaskedLM, AutoModelForTokenClassification
from transformers import BertTokenizer, BertModel, BertForMaskedLM, BertForQuestionAnswering, BertForPreTraining
from transformers import T5ForConditionalGeneration, MT5ForConditionalGeneration, PegasusForConditionalGeneration, PegasusTokenizer, T5Tokenizer
from transformers import GPT2LMHeadModel, GPT2Tokenizer
from transformers import RoFormerPreTrainedModel, RoFormerTokenizer, RoFormerForMaskedLM
from transformers import ElectraModel, ElectraTokenizer, ElectraForPreTraining
from transformers import XLNetLMHeadModel 
from transformers import LongformerModel
from transformers import BartForConditionalGeneration, BartTokenizer


ROOT = "/pretrains/pt"

@dataclass
class ModelCard:
    web_dirname : str                        # hfl/bert-base-wwm
    model : object = AutoModel
    tokenizer : object = AutoTokenizer
    save_dir : str = None                    # example: /pretrians/pt/hfl-bert-base-wwm

    @property
    def local_dirname(self):
        return self.web_dirname.replace("/", "-", 1)     # hfl-bert-base-wwm
    
    @property
    def web_url(self):
        return "https://huggingface.co/"+self.web_dirname

    @property
    def local_url(self):
        if self.save_dir is None:
            return os.path.join(ROOT, self.local_dirname)
        else:
            return self.save_dir

    def save(self):
        t1 = time.time()
        print(time.strftime("%Y-%m-%d %H:%M:%S", time.localtime(t1)))
        print("="*20, f"  start saving {self.web_url} !  ", "="*20)
        config = AutoConfig.from_pretrained(self.web_dirname)
        if self.model != AutoModel:
            self.m = self.model.from_pretrained(self.web_dirname)
        else:
            self.m = self.model.from_config(config)
        self.t = self.tokenizer.from_pretrained(self.web_dirname)
        # self.t = self.tokenizer(config)
        t2 = time.time()
        print(time.strftime("%Y-%m-%d %H:%M:%S", time.localtime(t2)), f"---- Downloading finished in {(t2-t1):.2f} seconds.")
        self.m.save_pretrained(self.local_url)
        self.t.save_pretrained(self.local_url)
        print(f"saved to {self.local_url}; please check if warning info exists \n "
                f"Or you'd better re-save the model with correct model head ! {self.web_url}")

    def is_saved(self):
        return os.path.exists(self.local_url)

models = [
    
    ModelCard('hfl/chinese-roberta-wwm-ext-large', model=BertForPreTraining),
    ModelCard('hfl/chinese-bert-wwm-ext', model=BertForPreTraining),
    ModelCard('hfl/chinese-bert-wwm', model=BertForPreTraining),
    ModelCard('hfl/chinese-macbert-base', model=BertForPreTraining),
    ModelCard('hfl/chinese-macbert-large', model=BertForPreTraining),
    ModelCard('hfl/chinese-electra-180g-large-discriminator', model=ElectraForPreTraining, tokenizer=ElectraTokenizer),
    ModelCard('hfl/chinese-legal-electra-large-discriminator', model=ElectraForPreTraining, tokenizer=ElectraTokenizer),
    ModelCard('hfl/chinese-xlnet-mid', model=XLNetLMHeadModel),
    ModelCard('hfl/chinese-xlnet-base', model=XLNetLMHeadModel),
    ModelCard('hfl/rbt3', model=BertForPreTraining),           # re-trained 3-layer RoBERTa-wwm-ext model
    # QA
    ModelCard('luhua/chinese_pretrain_mrc_roberta_wwm_ext_large', model=BertForQuestionAnswering, tokenizer=BertTokenizer),
    ModelCard('luhua/chinese_pretrain_mrc_macbert_large', model=BertForQuestionAnswering, tokenizer=BertTokenizer),
    ModelCard('uer/roberta-base-chinese-extractive-qa', model=BertForQuestionAnswering, tokenizer=BertTokenizer),
    # ModelCard('mrm8488/spanbert-large-finetuned-squadv2', model=BertForQuestionAnswering, tokenizer=BertTokenizer),
    ModelCard('schen/longformer-chinese-base-4096', model=LongformerModel, tokenizer=BertTokenizer),
    # https://github.com/CLUEbenchmark/CLUEPretrainedModels
    # ModelCard('clue/roberta_chinese_clue_tiny', model=BertModel),
    # ModelCard('clue/roberta_chinese_pair_tiny', model=BertModel, tokenizer=BertTokenizer),
    # ModelCard('clue/roberta_chinese_3L768_clue_tiny', model=BertModel),
    # ModelCard('clue/roberta_chinese_3L312_clue_tiny' , model=BertModel),
    # ModelCard('clue/roberta_chinese_clue_large' , model=BertModel),
    # ModelCard('clue/roberta_chinese_pair_large', model=BertModel),
    # ModelCard('clue/xlnet_chinese_large', model=XLNetLMHeadModel),
    # https://github.com/renmada/t5-pegasus-pytorch
    ModelCard('imxly/t5-pegasus',  model=T5ForConditionalGeneration, tokenizer=BertTokenizer),
    ModelCard('imxly/t5-pegasus-small',  model=T5ForConditionalGeneration, tokenizer=BertTokenizer),
    ModelCard('algolet/mt5-base-chinese-qg',  model=MT5ForConditionalGeneration, tokenizer=T5Tokenizer),
    ModelCard('IDEA-CCNL/Randeng-T5-77M',  model=MT5ForConditionalGeneration, tokenizer=T5Tokenizer),
    ModelCard('IDEA-CCNL/Randeng-T5-784M',  model=MT5ForConditionalGeneration, tokenizer=T5Tokenizer),
    ModelCard('IDEA-CCNL/Randeng-MegatronT5-770M',  model=MT5ForConditionalGeneration, tokenizer=T5Tokenizer),  # something error
    ModelCard('IDEA-CCNL/Randeng-Pegasus-238M-Summary-Chinese',  model=PegasusForConditionalGeneration, tokenizer=PegasusTokenizer),
    ModelCard('IDEA-CCNL/Randeng-Pegasus-523M-Summary-Chinese',  model=PegasusForConditionalGeneration, tokenizer=PegasusTokenizer),
    ModelCard('IDEA-CCNL/Randeng-Pegasus-523M-Chinese',  model=PegasusForConditionalGeneration, tokenizer=AutoTokenizer),
    ModelCard('IDEA-CCNL/Randeng-BART-139M',  model=BartForConditionalGeneration, tokenizer=AutoTokenizer),
    ModelCard('IDEA-CCNL/Randeng-BART-139M-SUMMARY',  model=BartForConditionalGeneration, tokenizer=AutoTokenizer),
    ModelCard('IDEA-CCNL/Wenzhong2.0-GPT2-3.5B',  model=GPT2LMHeadModel, tokenizer=GPT2Tokenizer),
    ModelCard('IDEA-CCNL/Wenzhong-GPT2-110M',  model=GPT2LMHeadModel, tokenizer=GPT2Tokenizer),
    ModelCard('Langboat/mengzi-t5-base',  model=T5ForConditionalGeneration, tokenizer=T5Tokenizer),

    # https://huggingface.co/uer
    ModelCard('uer/chinese_roberta_L-2_H-128', model=BertForMaskedLM, tokenizer=BertTokenizer),
    ModelCard('uer/chinese_roberta_L-2_H-768', model=BertForMaskedLM, tokenizer=BertTokenizer),
    ModelCard('uer/t5-base-chinese-cluecorpussmall',  model=T5ForConditionalGeneration, tokenizer=BertTokenizer),
    ModelCard('uer/t5-small-chinese-cluecorpussmall',  model=T5ForConditionalGeneration, tokenizer=BertTokenizer),
    ModelCard('uer/bart-base-chinese-cluecorpussmall', BartForConditionalGeneration, BertTokenizer),
    ModelCard('uer/gpt2-chinese-poem', model=GPT2LMHeadModel, tokenizer=BertTokenizer),
    ModelCard('uer/gpt2-base-chinese-cluecorpussmall', model=GPT2LMHeadModel, tokenizer=BertTokenizer),
    ModelCard('uer/gpt2-distil-chinese-cluecorpussmall', model=GPT2LMHeadModel, tokenizer=BertTokenizer),
    ModelCard('uer/pegasus-base-chinese-cluecorpussmall', model=PegasusForConditionalGeneration, tokenizer=BertTokenizer),

    ModelCard('voidful/albert_chinese_xxlarge', model=BertModel, tokenizer=BertTokenizer),
    # ModelCard('wptoux/albert-chinese-large-qa'),
    ModelCard("junnyu/roformer_chinese_small", model=RoFormerForMaskedLM, tokenizer=RoFormerTokenizer),
    ModelCard("junnyu/roformer_chinese_base", model=RoFormerForMaskedLM, tokenizer=RoFormerTokenizer),
    ModelCard("junnyu/roformer_chinese_char_base", model=RoFormerForMaskedLM, tokenizer=RoFormerTokenizer),
    ModelCard("ckiplab/gpt2-base-chinese", model=AutoModelForMaskedLM, tokenizer=BertTokenizer),     # https://huggingface.co/ckiplab/gpt2-base-chinese
    ModelCard('fnlp/bart-base-chinese', BartForConditionalGeneration, BertTokenizer),
    # 古文
    ModelCard('ethanyt/guwenbert-base'),         # https://huggingface.co/ethanyt/guwenbert-base
    ModelCard('ethanyt/guwenbert-large'),
    ModelCard('SIKU-BERT/sikubert'),
    ModelCard('SIKU-BERT/sikuroberta'),
    ModelCard('uer/gpt2-chinese-ancient', model=GPT2LMHeadModel, tokenizer=BertTokenizer),
    # English
    ModelCard('prajjwal1/bert-tiny'),
    ModelCard('bert-base-uncased'),
    ModelCard('bert-large-uncased'),
    ModelCard('xlm-roberta-large'),
    ModelCard('distilroberta-base'),
    ModelCard('gpt2-large', model=GPT2LMHeadModel),
    ModelCard('gpt2', model=GPT2LMHeadModel),
    ModelCard('distilgpt2', model=GPT2LMHeadModel),
    ModelCard('microsoft/DialoGPT-large'),
    ModelCard('facebook/bart-base', BartForConditionalGeneration, BartTokenizer),
    ModelCard('allenai/longformer-base-4096'),
    ModelCard('allenai/t5-small-squad11', model=T5ForConditionalGeneration),
    ModelCard('allenai/longformer-large-4096-finetuned-triviaqa'),
    ModelCard('allenai/unifiedqa-t5-small', model=T5ForConditionalGeneration),
    ModelCard('allenai/unifiedqa-t5-11b', model=T5ForConditionalGeneration),
    ModelCard('allenai/unifiedqa-t5-large', model=T5ForConditionalGeneration),
    ModelCard('allenai/t5-small-squad2-question-generation', model=T5ForConditionalGeneration),
    ModelCard('SpanBERT/spanbert-base-cased'),
    ModelCard('SpanBERT/spanbert-large-cased'),
    ModelCard('tuner007/pegasus_paraphrase', model=PegasusForConditionalGeneration, tokenizer=PegasusTokenizer),    #https://huggingface.co/tuner007/pegasus_paraphrase
    ModelCard('t5-large', model=T5ForConditionalGeneration),
    ModelCard('t5-small', model=T5ForConditionalGeneration),
    ModelCard('t5-base', model=T5ForConditionalGeneration),
    ModelCard('facebook/bart-large',BartForConditionalGeneration, BartTokenizer),
    # https://huggingface.co/sshleifer
    ModelCard('sshleifer/tinier_bart', BartForConditionalGeneration, BartTokenizer),
    ModelCard('sshleifer/distilbart-xsum-1-1', BartForConditionalGeneration, BartTokenizer),
    ModelCard('sshleifer/distilbart-xsum-12-1', BartForConditionalGeneration, BartTokenizer),
    ModelCard('sshleifer/distilbart-xsum-12-6', BartForConditionalGeneration, BartTokenizer),
    ModelCard('sshleifer/distilbart-xsum-6-6', BartForConditionalGeneration, BartTokenizer),

    ModelCard('google/bigbird-roberta-base'),
    ModelCard('google/bigbird-roberta-large'),
    ModelCard('google/byt5-small',model=T5ForConditionalGeneration,),
    ModelCard('google/byt5-base',model=T5ForConditionalGeneration),
    ModelCard('google/byt5-large',model=T5ForConditionalGeneration),
    ModelCard('google/byt5-xl',model=T5ForConditionalGeneration),
    ModelCard('google/byt5-xxl',model=T5ForConditionalGeneration),
    ModelCard('google/mt5-small',model=MT5ForConditionalGeneration),
    ModelCard('google/mt5-base' ,model=MT5ForConditionalGeneration),
    ModelCard('google/mt5-large',model=MT5ForConditionalGeneration),
    ModelCard('google/mt5-xxl',model=MT5ForConditionalGeneration),
    ModelCard('google/roberta2roberta_L-24_cnn_daily_mail'),
    ModelCard('google/bigbird-pegasus-large-arxiv'),
    # ModelCard(''),
] 


def get_model_url(string):
    """
    如果该预训练模型已经被保存到本地了，则返回本地路径
    否则，如果在本文件的模型集合中，则下载后保存，如果不在，则直接保存到本地； 最后返回本地路径
    Args:
        string (str): 预训练模型的huggingface标识， 如 hfl/roberta_chinese_wwm

    Returns:
        str : 本地路径
    """
    tmp_model = ModelCard(string)
    if tmp_model.is_saved():
        return tmp_model.local_url
    else:
        for model in models:
            if model.web_dirname == string:
                model.save()
                return model.local_url
            else:
                tmp_model.save()
                return tmp_model.local_url

def saved_info():
    print("\n".join([ f"{_.is_saved()}\t{i}\t{_.local_url}"  for i, _ in enumerate(models)]))

def download(url, model_cls=AutoModel, tokenizer_cls=AutoTokenizer, path="./hgf_model", **kwargs):
    """手动下载模型
    Args:
        url: _description_
        model_cls: _description_. Defaults to AutoModel.
        tokenizer_cls: _description_. Defaults to AutoTokenizer.
        path: _description_. Defaults to "./hgf_model".
    """
    model = model_cls.from_pretrained(url, **kwargs)
    tok = tokenizer_cls.from_pretrained(url)
    if not os.path.exists(path):
        os.makedirs(path)
    model.save_pretrained(path)
    tok.save_pretrained(path)
    print(f"{url} saving to {path}!")
    

def save_model(name, model, tokenizer):
    """
        huggingface model url
        model class
        model tokenizer class 
    """
    tmp_model = ModelCard(name, model, tokenizer)
    tmp_model.save()


def hf_download(repo_id, local_dir):
    """ 
        调用官方接口，将模型下载到本地目录
    """
    from huggingface_hub import snapshot_download
    snapshot_download(
        repo_id=repo_id, 
        allow_patterns=["*.md", "*.txt", "*.json", "*.bin", "*.model"],  # pytorch 
        ignore_patterns=None,
        local_dir=local_dir,
        local_dir_use_symlinks=False
    )


if __name__ == '__main__':
    # for _ in models:
        # print(_.local_url)
    saved_info()
    



from .common import * 


if __name__ == '__main__':
    PLM = "/pretrains/pt/fb_bart"
    DATA = ""
    TOKENIZE_FN = None   # from .kqa import nl2cy_tokenization
    from qdls.data import load_json
    from qdls.utils import print_dict
    from qdls.reg.collators.common_collators import Seq2seqCollator
    
    import datasets 
    from transformers import AutoTokenizer
    from torch.utils.data import DataLoader
    tokenizer = AutoTokenizer.from_pretrained(PLM)
    tokenizer.pad_token_id = 0 
    

    data = load_json(DATA)[:100]
    for sample in data:
        r = TOKENIZE_FN(example=sample, tokenizer=tokenizer)
        print_dict(r)
        print(tokenizer.convert_ids_to_tokens(r['distill_labels']))
        input()
        break
    ds = datasets.Dataset.from_list(data)
    ds = ds.map(TOKENIZE_FN, num_proc=1, fn_kwargs={'tokenizer':tokenizer})
    loader = DataLoader(ds, batch_size=4, collate_fn=Seq2seqCollator(tokenizer))
    for batch in loader:
        print(batch)
        # print(tokenizer.batch_decode(batch['labels'], skip_special_tokens=False, clean_up_tokenization_spaces=False))
        # print(tokenizer.batch_decode(batch['distill_labels'], skip_special_tokens=False, clean_up_tokenization_spaces=False))
        import pdb;pdb.set_trace();
        break



# -*- coding: utf-8 -*-
# @File    :   __init__.py
# @Time    :   2023/11/22 15:23:19
# @Author  :   Qing 
# @Email   :   aqsz2526@outlook.com
######################### docstring ########################
'''
    声明 
        normalize_sample
        evaluate
    两个方法
'''

from .base import BaseEvalutor
from .kqa import KqaAutoEvaluator



from antlr4.tree.Tree import TerminalNodeImpl
from nltk import Tree


def parse_layer(tree, rule_names, parts, parents, indent = 0):
    """ 遍历树，保存叶子结点 """
    if tree.getText() == "<EOF>":
        return
    elif isinstance(tree, TerminalNodeImpl):
        # print("{0}TOKEN='{1}'".format("  " * indent, tree.getText()))
        # layer[indent].append(tree.getText())
        parts.append(tree.getText())
        parents.append(tree.getParent())
    elif tree.children is None:
        # print("none children", tree)
        return
    else:
        # print("{0}{1}".format("  " * indent, rule_names[tree.getRuleIndex()]))
        for child in tree.children:
            parse_layer(child, rule_names, parts, parents, indent + 1)



def save_ast(tree, parser, path):
    """ 
        将 antlr tree 保存为 png or pdf 
        ```
        tree, parser = parse_cypher(hop1['cypher'], True)
        save_ast(tree, parser, "hop1_cypher.pdf")
        ```
    """
    try:
        import svgling, cairosvg
    except Exception as e:
        print("Try ` pip install svgling cairosvg `")
        raise e 

    tree_string = tree.toStringTree(recog=parser)
    img = svgling.draw_tree(Tree.fromstring(tree_string))
    svg = img.get_svg()
    if path.endswith("pdf"):
        cairosvg.svg2pdf(svg.tostring(),  write_to=path, unsafe=False)
    elif path.endswith("png"):
        cairosvg.svg2png(svg.tostring(),  write_to=path)
    else:
        raise Exception(path, "is not pdf or png")


# -*- coding: utf-8 -*-
# @File    :   example.py
# @Time    :   2023/11/23 13:46:39
# @Author  :   Qing 
# @Email   :   aqsz2526@outlook.com
######################### docstring ########################
'''
    to use the evaluator, you need to:
    implement the following functions:
        normalize_sample
        evaluate
    
'''
from qdls.data import load_json, save_json, load_jsonl
from argparse import Namespace
from evaluator import BaseEvalutor, KqaAutoEvaluator


class KqaSparqlICLEvaluator(KqaAutoEvaluator):

    @staticmethod
    def _process_vllm(query):
        return query.split("</s>")[0].strip()

    def normalize_sample(self, sample, lang='sparql'):
        d = {
            'sample_id': sample['sample_id'],
            'question': sample['question'],
            'answer': sample['answer'],
            'sparql': sample['sparql'],
            'pred': KqaSparqlICLEvaluator._process_vllm(sample['vllm']),
        }
        return d
    
    def evaluate(self):
        scores = obj.calc_marco_metrics()
        for k,v in scores.items():
            print(f"{k:<20}: {v:.4f}")

class KqaCypherICLEvaluator(KqaAutoEvaluator):

    @staticmethod
    def _process_vllm(query):
        return query.split("</s>")[0].strip()
    

    def normalize_sample(self, sample, lang='cypher'):
        d = {
            'sample_id': sample['sample_id'],
            'question': sample['question'],
            'answer': sample['answer'],
            'cypher': sample.get("cypher") if "cypher" in sample else sample['match_clause'],
            'pred': KqaCypherICLEvaluator._process_vllm(sample['vllm']),
        } 
        return d

    def evaluate(self):
        scores = self.calc_marco_metrics()
        for k,v in scores.items():
            print(f"{k:<20}: {v:.4f}")


class KqaCypherICLEvaluatorGPT(KqaAutoEvaluator):
    """ 
    """
    def normalize_sample(self, sample, lang=None):
        d = {
            'sample_id': sample['sample_id'],
            'question': sample['question'],
            'answer': sample['answer'],
            'cypher': sample.get("cypher"),
            'pred': sample['gpt-3.5-turbo-0613'],
        } 
        return d
    
    def evaluate(self):
        scores = self.calc_marco_metrics()
        for k,v in scores.items():
            print(f"{k:<20}: {v:.4f}")


if __name__ == '__main__':
    # data = load_json("/home/qing/raid/paperwork/ICL/src/cache/results/kqa_qwen14b_BM25KqaRunner_5icl.jsonl")
    # obj = KqaAutoEvaluator(data, lang='sparql', nproc=16)

    # data = load_jsonl("/home/qing/raid/paperwork/ICL/src/cache/results/kqa_qwen14b_MixKqaSparqlRunner_6icl.jsonl")
    """ 
    bleu                : 0.8521
    executable          : 0.9369
    exact_match         : 0.4193
    is_correct          : 0.5158
    ./eval_results_for_sparql.json saved with 11797 samples!
    """

    # data = load_jsonl("/home/qing/raid/paperwork/ICL/src/cache/results/kqa_qwen14b_BM25KqaRunner_3icl.jsonl")
    """ 
    bleu                : 0.8439
    executable          : 0.9164
    exact_match         : 0.4354
    is_correct          : 0.5224
    ./eval_results_for_sparql.json saved with 11797 samples!
    """

    # data = load_jsonl("/home/qing/raid/paperwork/ICL/src/cache/results/kqa_qwen14b_VectorKqaRunner_6icl.jsonl")
    """ 
    bleu                : 0.8819
    executable          : 0.9606
    exact_match         : 0.4608
    is_correct          : 0.5605
    """
    # neo4j_config = Namespace(
    #     neo4j_uri="neo4j://map:28892", neo4j_user="neo4j", neo4j_passwd="kqa", timeout=3
    # )

    # obj = KqaSparqlICLEvaluator(data, lang='sparql', nproc=32, neo4j_config=neo4j_config)
    # obj.neo4j_config = neo4j_config
    # obj.evaluate()
    # obj.save_processed()
    
    def _chatgpt_icl_example():

        neo4j_config = Namespace(
            neo4j_uri="neo4j://map:28892", neo4j_user="neo4j", neo4j_passwd="kqa", timeout=3
        )
        data = load_json("/home/qing/raid/paperwork/manage/chatgpt/cache/nl2cql_1k.json")
        obj = KqaCypherICLEvaluatorGPT(data, lang='cypher', nproc=32, neo4j_config=neo4j_config, metrics_to_calc=['exact_set_match'])
        obj.evaluate()
        obj.save_processed("/home/qing/raid/paperwork/manage/chatgpt/cache/nl2cql_1k_processed.json")
        """ 
        bleu                : 0.6302
        executable          : 0.9130
        exact_match         : 0.0230
        is_correct          : 0.1450
        exact_set_match     : 0.0520
        """

    _chatgpt_icl_example()



import os 
try:
    import faiss
except ImportError:
    raise ImportError("Please install faiss by running `pip install faiss-gpu`")
from loguru import logger
try:
    from sentence_transformers import SentenceTransformer  
    from langchain_huggingface import HuggingFaceEmbeddings  # requires sentence-transformers 
except ImportError:
    raise ImportError("Please install langchain-huggingface by running `pip install langchain-huggingface`")

try:
    from langchain_chroma import Chroma 
    from langchain_core.documents import Document
except ImportError:
    logger.warning("Please install langchain-chroma by running `pip install langchain-chroma`\nOr use VectorRetriever backended by faiss")

from qdls.llm_infer.retrievers.base_retriever import Retriever

class VectorRetriever(Retriever):
    def __init__(self, data_source, key=None, embedding_model_path=None, cache_file=None, **kwargs):
        """
        Args:
            data_source: List[Dict] or List[Object].
            key: str or callable function. fetch data from each sample in data_source.
            embedding_model_path: model used to encode text. 
            cache_file: file path to save the cached Vector/Index

        """
        self.data_source = data_source
        assert key is not None, "key is not specified"
        self.cache_file = cache_file      
        if type(key) is str:
            self.items4retrieval = [ sample[key] for sample in data_source]
        elif callable(key):
            # 以构建更多样的检索形式，而非简单的使用sample[key]作为检索项
            self.items4retrieval = [ key(sample) for sample in data_source]
        else:
            raise Exception(f"key should be dict key or callable function")
        self.key = key 
        self.model = self._init_encoding_model(embedding_model_path)

        # 
        self._handle_cache( self.encode_fn if hasattr(self, 'encode_fn') else None)

        # default args 
        self.bsz = kwargs.get('bsz', 128)
    

    def _init_encoding_model(self, model_name_or_path=None):
        if model_name_or_path is None:
            model_name_or_path = "/sshfs/pretrains/sentence-transformers/all-MiniLM-L6-v2"
            # option models: /sshfs/pretrains/Salesforce/SFR-Embedding-Mistral
        model = SentenceTransformer(model_name_or_path, trust_remote_code=True)
        return model 
            
    def _handle_cache(self, encode_fn=None):
        os.makedirs("cached_index", exist_ok=True)
        if self.cache_file is not None:
            cache_file = os.path.join(".", "cached_index",f"{self.cache_file}_index.bin")
        else:
            cache_file = f"{ self.key if type(self.key) is str else self.key.__name__ }_index.bin"
            cache_file = os.path.join(".", "cached_index", cache_file)

        if encode_fn is None:
            def encode_fn(model, text, bsz=1):
                return model.encode(text, batch_size=bsz, show_progress_bar=True)

        if os.path.exists(cache_file):
            self.index = faiss.read_index(cache_file)
            logger.info(f"index loaded from {cache_file}")
        else:
            # vectors = self.model.encode(self.items4retrieval, batch_size=128, show_progress_bar=True)
            vectors = encode_fn(self.model, self.items4retrieval)
            self.index = faiss.IndexFlatIP(vectors.shape[-1])
            logger.info(f"index built, dimension: {self.index.d} vs {vectors.shape}, ntotal: {self.index.ntotal}")
            self.index.add(vectors)
            faiss.write_index(self.index, cache_file)
            logger.info(f"index built and saved to {cache_file}")

    
    def get_topk_samples(self, query, topk=5):
        """ 将输入query分词，返回self.topk个 sample """
        if type(query) is str:
            query_vector = self.model.encode(query)
            D, I = self.index.search(query_vector.reshape(1,-1), topk)
        else:
            raise Exception("query should be str")
        topk_samples = [ self.data_source[i] for i in I[0]]
        return topk_samples


class VectorRetrieverInstruct(VectorRetriever):
    """ 
        一些模型是带着 instruction 训练的
        ref: https://huggingface.co/Salesforce/SFR-Embedding-Mistral
    """
    def encode_fn(self, model, text, bsz):
        task_description = 'Given a web search query, retrieve relevant passages that answer the query'
        text = [f'Instruct: {task_description}\nQuery: {query}' for query in text]
        return model.encode(text, batch_size=bsz, show_progress_bar=True)


class VectorRetrieverLangChain(VectorRetriever):

    def __init__(self, data_source, key=None, embedding_model_path=None, cache_file=None):
        """
        Args:
            data_source: List[Dict] or List[Object].
            key: str or callable function. fetch data from each sample in data_source.
            embedding_model_path: model used to encode text. 
            cache_file: file path to save the cached Vector/Index

        """
        self.data_source = data_source
        assert key is not None, "key is not specified"
        self.cache_file = cache_file
        if type(key) is str:
            # use metadata property to store the original sample
            self.items4retrieval = [Document(sample[key], metadata=sample) for sample in data_source]
        elif callable(key):
            # 以构建更多样的检索形式，而非简单的使用sample[key]作为检索项
            self.items4retrieval = [Document(key(sample), metadata=sample) for sample in data_source]
        else:
            raise Exception(f"key should be dict key or callable function")
        self.key = key 
        self.model = self._init_encoding_model(embedding_model_path)
        self._handle_cache()


    def _init_encoding_model(self, model_name_or_path=None):
        if model_name_or_path is None:
            model_name_or_path = "/sshfs/pretrains/BAAI/bge-large-zh-v1.5"
        model_kwargs = {'device': 'cuda',  }
        encode_kwargs = {'normalize_embeddings': True}    
        model = HuggingFaceEmbeddings(
                    model_name=model_name_or_path,
                    model_kwargs=model_kwargs,
                    encode_kwargs=encode_kwargs,
                    show_progress=True,
                )
        return model
    
    def _handle_cache(self):
        os.makedirs("cached_index", exist_ok=True)
        if self.cache_file is not None:
            persist_directory = self.cache_file + '_index'
        else:
            persist_directory = f"{ self.key if type(self.key) is str else self.key.__name__ }_index"
        persist_directory = os.path.join(".", "cached_index", persist_directory)
        

        collection_name = "default"
        
        if os.path.exists(persist_directory):
            vectorstore = Chroma(
                collection_name=collection_name,
                embedding_function=self.model,
                persist_directory=persist_directory,
            )
            logger.info(f"Vectorstore loaded from {persist_directory}")
        else:
            chunks = self.items4retrieval
            vectorstore = Chroma.from_documents(
                documents=chunks, 
                collection_name=collection_name,
                embedding=self.model,
                persist_directory=persist_directory,
            )
            logger.info(f"Vectorstore built and saved to {persist_directory}") 
        
        self.retriever = vectorstore

    def get_topk_samples(self, query:str, topk=5):

        r = self.retriever.similarity_search(query, k=topk)
        return [ doc.metadata for doc in r]


if __name__ == '__main__':
    data = [
        {"id": 1, "text": "今天天气不错"},
        {"id": 2, "text": "今天天气不好"},
        {"id": 3, "text": "今天天气很差"},
    ]
    # r = VectorRetrieverLangChain(data_source=data, key="text")
    # r = VectorRetriever(data_source=data, key="text", embedding_model_path="/sshfs/pretrains/Salesforce/SFR-Embedding-Mistral", bsz=2)
    r = VectorRetriever(data_source=data, key="text", embedding_model_path="/sshfs/pretrains/Alibaba-NLP/gte-Qwen2-7B-instruct", bsz=2)
    query = "坏天气"
    print(r.get_topk_samples(query, topk=1))


# -*- coding: utf-8 -*-
# @File    :   register.py
# @Time    :   2023/05/22 16:11:26
# @Author  :   Qing 
# @Email   :   aqsz2526@outlook.com
######################### docstring ########################
'''
    注册器的实现, ref to https://github.com/Delta-ML/delta/blob/master/delta/utils/register.py

    model:                  一些常用的模型实现，默认不加入注册器 only_data=False
    collator:               collator的实现
    process_function:       数据集的输入构建、分词处理
    datamodule:             pytorch lightning 的 datamodule 实现
'''
import importlib
import logging
import os 
import sys 
FILE_DIR = os.path.dirname(os.path.abspath(__file__))

############# 增加新的文件之后需要修改下面的module列表 #############
# 文件夹下面的 py 文件名

MODEL_MODULES = [
    "test_model"
]

COLLATE_MODULES = [
   "common_collators",
]

PROCESS_MODULES = [
   "kqa"
]

DM_MODULES = [
    "base_dm"
]

ALL_DATA_MODULES = [
    ("qdls.reg.collators", COLLATE_MODULES),
    ("qdls.reg.process_functions", PROCESS_MODULES),
    ("qdls.reg.datamodules", DM_MODULES),
]

ALL_MODULES = [
   ("qdls.reg.models", MODEL_MODULES), 
] + ALL_DATA_MODULES

########## 以下内容不需要修改 ##########
 
class Register:

    def __init__(self, registry_name):
        self._dict = {}
        self._name = registry_name

    def __setitem__(self, key, value):
        """ 取函数名作为key """
        if not callable(value):
            raise Exception(f"Value of a Registry must be a callable!\nValue: {value}")
        if key is None:
            key = value.__name__
        if key in self._dict:
            logging.warning("Key %s already in registry %s." % (key, self._name))
        self._dict[key] = value

    def register(self, target):
        """Decorator to register a function or class."""

        def add(key, value):
            self[key] = value
            return value

        if callable(target):
            # @reg.register
            return add(None, target)
        # @reg.register('alias')
        return lambda x: add(target, x)

    def __getitem__(self, key):
        return self._dict[key]

    def __contains__(self, key):
        return key in self._dict

    def keys(self):
        """key"""
        return self._dict.keys()

    def get(self, key):
        """ get a data process function from registry"""
        return self._dict[key]


class registers:
    model = Register('model')
    collator = Register('collator')
    process_function = Register('process_function')
    datamodule = Register('datamodule')
    
    def __init__(self) -> None:
        raise RuntimeError("Registries is not intended to be instantiated")
    

def _handle_errors(errors):
  """Log out and possibly reraise errors during import."""
  if not errors:
    return
  for name, err in errors:
    logging.warning("Module {} import failed: {}".format(name, err))
  logging.fatal("Please check these modules.")


def path_to_module_format(py_path):
  """Transform a python file path to module format."""
  return py_path.replace("/", ".").rstrip(".py")


def add_custom_modules(all_modules, config=None):
  """Add custom modules to all_modules"""
  current_work_dir = os.getcwd()
  if current_work_dir not in sys.path:
    sys.path.append(current_work_dir)
  if config is not None and "custom_modules" in config:
    custom_modules = config["custom_modules"]
    if not isinstance(custom_modules, list):
      custom_modules = [custom_modules]
    all_modules += [
        ("", [path_to_module_format(module)]) for module in custom_modules
    ]


def import_all_modules_for_register(config=None, only_data=True):
  """Import all modules for register."""
  if only_data:
    all_modules = ALL_DATA_MODULES
  else:
    all_modules = ALL_MODULES

  add_custom_modules(all_modules, config)

  logging.debug(f"All modules: {all_modules}")
  errors = []
  for base_dir, modules in all_modules:
    for name in modules:
      try:
        if base_dir != "":
          full_name = base_dir + "." + name
        else:
          full_name = name
        # logging.warning(f"Importing {full_name}...")
        importlib.import_module(full_name)
        logging.debug(f"{full_name} loaded.")
      except ImportError as error:
        errors.append((name, error))
  _handle_errors(errors)





# -*- coding: utf-8 -*-
# @File    :   metric_fns.py
# @Time    :   2023/11/20 14:05:24
# @Author  :   Qing 
# @Email   :   aqsz2526@outlook.com
######################### docstring ########################
'''
    用于评估 GQL 生成模型的指标函数
'''
from qdls.reg.register import Register
from qdls.gql.cypher.utils.syntax import syntax_check as syntax_check_cypher
from qdls.gql.sparql.utils.syntax import syntax_check as syntax_check_sparql

from qdls.gql.cypher.utils.kqa_eval import exec_one_sample_cypher
try:
    from torchmetrics.text import sacre_bleu_score # new in torchmetrics 2.0
except:
    from torchmetrics.functional import sacre_bleu_score

from qdls.kgqa_eval.exact_set.evaluation_em import evaluate_em

from pygments.lexers import get_lexer_by_name
from pygments.lexers.rdf import SparqlLexer
from pygments.token import *

sp_lexer = SparqlLexer()
cy_lexer = get_lexer_by_name('py2neo.cypher') # pip install py2neo to fix

metric_fns = Register('metric_fns')


@metric_fns.register('exact_set_match')
def exact_set_match(pred, gold, verbose=False):
    """ 
    调用 另外实现的 exact set match 
    比较的是几个子句中的字符串
    """
    try:
        is_match, gold_splits, pred_splits, dismatch_gold, dismatch_pred = evaluate_em(gold, pred)
    except Exception as e:
        return 0.0
    return 1.0 if is_match else 0.0


@metric_fns.register("exact_match")
def match_query(pred, gold, lexer='cypher', verbose=False):
    """ 将cypher or sparql 语言解析为语义单元，比较是否匹配
        如果出现ERROR tag 则表示生成的语法错误， return 0 
    """
    lexer = sp_lexer if lexer == 'sparql' else cy_lexer
    pred_units = []
    try:
        for tag, substr in lexer.get_tokens(pred):
            if tag is Token.Error:
                return 0
            # print(tag, substr)
            if substr.strip() != "":
                pred_units.append(substr)
    except Exception as e:
        return 0 # 语法错误,无法lex
    gold_units = [ substr for tag, substr in lexer.get_tokens(gold) if substr.strip() != "" ]
    res = (gold_units == pred_units)
    if res:
        return 1.0
    else:
        if verbose:
            print(f'{" ".join(gold_units)} \n {" ".join(pred_units)}')
        return 0.0
    

def executable_fn_cypher(pred):
    """ 判断预测的是否可执行 """
    try:
        flag, tree, parser = syntax_check_cypher(pred)
    except:
        flag = False
    return 1.0 if flag else 0.0


def executable_fn_sparql(pred):
    """ 判断预测的是否可执行 """
    try:
        flag, tree, parser = syntax_check_sparql(pred)
    except:
        flag = False
    return 1.0 if flag else 0.0

@metric_fns.register("executable")
def executable_fn(pred, lang):
    if lang == 'sparql':
        return executable_fn_sparql(pred)
    elif lang in ['cql', 'cypher']:
        return executable_fn_cypher(pred) 
    else:
        raise Exception(f"{lang} is not supported!")


@metric_fns.register("bleu")
def bleu_fn(pred, refs):
    """ 默认是用 mteval-v13a 进行分词 return bleu-4 """
    return sacre_bleu_score([pred], [refs], tokenize='char').item()


def calc_metrics_per_sample(sample, ref_key, metrics=None, neo4j_config=None):
    """ 默认的预测 key 为 `pred`， 参考答案的key为 ref_key 
        sample normalized_sample()的结果
        ref_key 为参考答案的key: cypher,cql, sparql
        metrics 为要计算的指标， 默认为 ['bleu', 'executable', 'exact_match']
        neo4j_config 为neo4j的配置 

        新增 metric:  
        ``` from qdls.kgqa_eval.evaluator import metric_fns
            @metric_fns.register("NEW_METRIC")
            def NEW_METRIC_fn(sample, **kwargs): # kwargs 用于传递额外参数
                return res 
        ``` 
        最终效果: `sample['NEW_METRIC'] = NEW_METRIC_fn(sample)`
        如果这个 function 返回的是一个 dict，则会将其 update 到 sample 中
    """
    if metrics is None:
        metrics = ['bleu', 'executable', 'exact_match']
    else:
        assert isinstance(metrics, list), f"metrics should be a list of str, but got f{metrics}"

    if 'bleu' in metrics:
        bleu = bleu_fn(sample['pred'], [sample[ref_key]])
        sample['bleu'] = bleu
    
    if 'executable' in metrics:
        sample['executable'] = executable_fn(sample['pred'], ref_key)

    if 'exact_match' in metrics:
        sample['exact_match'] = match_query(sample['pred'], sample[ref_key], lexer=ref_key)

    if 'exact_set_match' in metrics:
        sample['exact_set_match'] = exact_set_match(sample['pred'], sample[ref_key])

    if 'is_correct' in metrics or 'exec_info' in metrics:
        assert ref_key in ['cql', 'cypher'], f"only cql or cypher is supported, but got {ref_key}"
        try:
            is_correct, info, results = exec_one_sample_cypher(sample, neo4j_config, key='pred')
        except Exception as e:
            print(e)
            is_correct, info, results = False, "Timeout", '[]'
        sample['is_correct'] = is_correct
        sample['exec_info'] = info
        sample['exec_results'] = results

    # 处理其他 metric 
    for k in metrics:
        if k in sample:
            continue
        fn = metric_fns.get(k) # 获取注册的函数
        res = fn(sample, neo4j_config=neo4j_config, ref_key=ref_key)
        if type(res) is dict:
            sample.update(res)
        else:
            sample[k] = res 

    # import pdb;pdb.set_trace();
    return sample

def nltk_bleu():
    import nltk
    reference = ['今天天气很好吗']

    # 候选翻译（机器翻译结果）
    candidate = '今日天气很好。'

    # 将参考翻译和候选翻译转换为标记化的句子列表
    reference_tokens = nltk.word_tokenize(reference[0], language='chinese')
    candidate_tokens = nltk.word_tokenize(candidate, language='chinese')

    # 计算BLEU分数
    bleu_score = nltk.translate.bleu_score.sentence_bleu([reference_tokens], candidate_tokens)

    print("BLEU score:", bleu_score)

if __name__ == '__main__':
    # print(bleu_fn('hello world', ['hello world']))
    nltk_bleu()



from ..grammar.CypherLexer import CypherLexer
from ..grammar.CypherListener import CypherListener
from ..grammar.CypherParser import CypherParser

from antlr4 import InputStream,CommonTokenStream
from antlr4.error.ErrorListener import ErrorListener


class MyParserErrorListener(ErrorListener):
    def syntaxError(self, recognizer, offendingSymbol, line, column, msg, e):
        raise Exception(f"Syntax Error:{msg}\n{e}\n{column}\n{line}\n{offendingSymbol}")

class MyLexerErrorListener(ErrorListener):
    def syntaxError(self, recognizer, offendingSymbol, line, column, msg, e):
        raise Exception(f"Lexer Error:{msg}\n{e}\n{column}\n{line}\n{offendingSymbol}")

def syntax_check(cypher):
    """
        using antlr4 to pre-check if generated cypher is syntax correct
        using self-defined ErrorListener to avoid console output
    """
    input_stream = InputStream(cypher)
    lexer = CypherLexer(input_stream)
    lexer._listeners = []
    lexer.addErrorListener(MyLexerErrorListener())
    stream = CommonTokenStream(lexer)
    parser = CypherParser(stream)
    parser._listeners = []
    parser.addErrorListener(MyParserErrorListener())
    try:
        tree = parser.oC_Cypher()
    except Exception as e:
        # print(e)
        return False, e, parser
    return True, tree, parser


if __name__ == '__main__':
    
    s1 = "match (q:SS return q limit 1"

    flag, _ = syntax_check(s1)

    print(flag)
    print("error msg:", _)

    """
    Syntax Error:no viable alternative at input 'match (q:SS return'
    None
    12
    1
    [@7,12:17='return',<62>,1:12]

    根据offendingSymbol可以定位到错误
    """



# -*- coding: utf-8 -*-
# @File    :   utils.py
# @Time    :   2023/05/21 16:47:34
# @Author  :   Qing 
# @Email   :   aqsz2526@outlook.com
######################### docstring ########################
'''
    常用的，不依赖于不常见模块的函数（使用时直接from qdls.utils import *）
'''
import json 
import sys
import omegaconf

from termcolor import colored, cprint
from argparse import Namespace

def print_dict(d):
    s = json.dumps(d, indent=4, ensure_ascii=False)
    print(s)


def build_config(d):
    """ 嵌套的 argparse.Namespace """
    config = Namespace()
    for k,v in d.items():
        if type(v) == dict:
            setattr(config,k, build_config(v))
        else:
            setattr(config, k, v)
    return config 


def config2dict(config):
    """ 将 argparse.Namespace 转换为 dict
        TODO: 支持更多类 dataclass 等 
    """
    if type(config) is omegaconf.dictconfig.DictConfig:
        d = omegaconf.OmegaConf.to_object(config)
        return d
    d = {}
    for k,v in config.__dict__.items():
        if type(v) is Namespace:
            d[k] = config2dict(v)
        else:
            d[k] = v 
    return d 


def print_string(string, color="yellow"):
    """ 在命令行以黄色字体打印 string 
        eg: ====================  string  ====================
    """
    if type(string) is not str:
        try:
            string = str(string)
        except:
            raise Exception("Input must be str or can be converted to str")

    s = "="*20 + string.center(len(string)+10,) + "="*20
    cprint(s, color, attrs=["bold"])


def print_config(config, title=None):
    """在命令行以 json 格式打印 config 绿色字体;   

    Args:
        config:  Config object  
        title: config 的名字. Defaults to None.
    """
    if title is None:
        print("="*80)
    else:
        print_string(title)

    config_dict = config2dict(config)
    config_string = json.dumps(config_dict, indent=4, ensure_ascii=False)
    cprint(config_string, "green")

    print_string("END OF CONFIG")


def namespace2dict(config):
    """ 将 argparse.Namespace 转换为 dict
    """
    d = {}
    for k,v in config.__dict__.items():
        if type(v) is Namespace:
            d[k] = namespace2dict(v)
        else:
            d[k] = v 
    return d 









def is_sample_valid(sample):
    """ span 抽取问题（SQuAD格式）,数据标注是否正确
    Args:
        sample (SquadSample): 读入的单条样本对象
    Returns:
        bool: 该数据的标注是否正确
    """
    if sample.answer_texts != []:
        text = sample.answer_texts[0]
        start_char = sample.answer_starts[0]
        # 标注的start_char的位置有问题
        if not sample.context[start_char:start_char+len(text)] == text:
            return False
    else:
        if sample.answer_starts != []:
            return False
    return True


# -*- coding: utf-8 -*-
# @File    :   stats.py
# @Time    :   2023/04/03 14:31:17
# @Author  :   Qing 
# @Email   :   aqsz2526@outlook.com
######################### docstring ########################
'''

'''

import numpy as np
import matplotlib.pyplot as plt
import matplotlib.pyplot as plt
from matplotlib.ticker import PercentFormatter

def draw_hist_percentage(numbers, nbins=50, title=None, save_file=None):
    """直方图 分成nbins个区间, y轴是每个区间数量的百分比

    Args:
        numbers: list 一组数字
        nbins: Int. Defaults to 50.
        title: Str. 图片的标题
        save_file: Str. filepath for saving the plot. Defaults to None.
    """
    fig, ax = plt.subplots(1, 1)
    # Plot the distribution using a histogram
    ax.hist(numbers, bins=nbins)
    ax.yaxis.set_major_formatter(PercentFormatter(len(numbers)))
    # Add labels and title
    plt.xlabel('Value')
    plt.ylabel('Percentage')
    title = title if title is not None else 'Distribution of numbers'
    plt.title(title)

    # Save the plot as a PNG file
    if save_file is not None:
        plt.savefig('logicv2_distr.png')


def test():
    numbers = np.random.randn(1000)
    draw_hist_percentage(numbers, nbins=50, title='Distribution of numbers', save_file=None)
    plt.show()


if __name__ == '__main__':
    test()



from qdls.data import load_json
import datasets 


class DatasetBuilder():
    """ 
        data:
            train_path  : local file path 
            val_path   : local file path
            test_path  : local file path
            dataset_name: huggingface dataset name
    """
    def __init__(self, config) -> None:
        """
           应当支持初始化两种方式
            1.huggingface 在线dataset
            2.本地文件 train_path val_path test_path
        """
        if not hasattr(config.data, 'dataset_name'):
            assert hasattr(config.data, 'train_path') and \
                    hasattr(config.data, 'val_path') and \
                    hasattr(config.data, 'test_path'), \
            f"train_path, val_path, test_path should be provided when dataset_name is not provided"
            self._init_from_paths(config)
        else:
            assert hasattr(config.data, 'dataset_name'), \
            f"you should provide dataset_name since no train_path, val_path, test_path is provided"
            self._init_from_hf(config)

        self.config = config 
        if hasattr(config.data, 'num_proc'):
            self.num_proc = config.data.num_proc
        else:
            self.num_proc = 4

    
    def _init_from_paths(self, config):
        """ 从本地文件加载数据, 允许'' """
        self.trainset = load_json(config.data.train_path) if config.data.train_path else []
        self.valset = load_json(config.data.val_path) if config.data.val_path else []
        self.testset = load_json(config.data.test_path) if config.data.test_path else []

        
    def _init_from_hf(self, config, proportion=0.8):
        """ 如果只有两个字段（大概率train test）则将train划分为train val """
        ds = datasets.load_dataset(config.data.dataset_name)
        if len(ds) == 2:
            train_ds = ds['train'].train_test_split(test_size=1-proportion)
            self.trainset = train_ds['train']
            self.valset = train_ds['test']

        self.testset = ds['test'] 
 
    
    def build(self, mode, tokenizer, tokenize_fn):
        """ 
            mode: train, val, test
        """
        if mode == 'train':
            data  = self.trainset
        elif mode == 'val':
            data  = self.valset 
        elif mode == 'test':
            data  = self.testset
        else:
            raise Exception(f"{mode} is not implemented!")
        
        if data == []:      # 对应的 trainset or valset or testset 为空
            return None 

        if type(data) is list:                              # 本地加载的数据是list
            dataset = datasets.Dataset.from_list(data)
        elif type(data) is datasets.Dataset:
            dataset = data 
        else:
            dataset = datasets.Dataset.from_dict(data)
        dataset = dataset.map(
            tokenize_fn, 
            batched=False,
            num_proc=self.num_proc, 
            fn_kwargs={'tokenizer': tokenizer, 'mode': mode}
        )
        return dataset
    
    def data_stat(self):

        pass 


import re


def match2where(cypher):
    cypher = cypher.replace(' ', '')
    pattern = re.compile(r'\([^)]*\)')
    matches = pattern.findall(cypher)
    wheres = []
    for match in matches:
        if '{' in match and '}' in match and ':' in match:
            wheres.append(match)
        elif '{' in match and '}' not in match and ':' in match:
            wheres.append(match + '}')
    match2where = []
    for where in wheres:
        tag = where.split(':')[1].split('{')[0]
        prop = where.split('{')[1].split(':')[0]
        value = where.split('{')[1].split(':')[1].split('}')[0]
        temp = tag + '.' + prop + ' == ' + value
        match2where.append(temp)
    return match2where


def special_match_find(cypher):
    pattern = re.compile(r'\{.*?\}')
    cypher = re.sub(pattern, '', cypher)
    return cypher


def find_matches(text):
    match_where = match2where(text)  # 取出match语句中的{name: '[s]'}特殊表示方式，并转为where条件
    text = special_match_find(text)  # 删除match语句中的{name: '[s]'}特殊表示方式
    text = re.sub(r'\s+', ' ', text)  # 去除多个连续空格
    text = text.lower().replace('match ', '')  # 去除match字符
    matches = re.findall(r':[^\)\]]+', text)
    # print(matches)
    matches = sorted([i.replace(' ', '')[1:] for i in matches])
    # print(matches)
    splits = {
        'match_tags': matches,
        'match_where': match_where
    }
    return splits


if __name__ == '__main__':
    # gold = 'MATCH (t:trade)<-[bt:belong_to]-(s:stock)-[hsd:has_stock_data]->(sd:stock_data)'
    # gold = 'MATCH (pof:public_offering_fund)'
    # gold = 'MATCH (s:stock)<-[ico:is_chairman_of]-(cm:chairman)'
    # gold = 'MATCH (s:stock)-[bt:belong_to]->(t:trade)'
    # gold = 'MATCH (sd:stock_data)<-[hsd:has_stock_data]-(s:stock)-[bt:belong_to]->(t:trade)'
    gold = "MATCH (t:trade)<-[bt:belong_to]-(s:stock{name: '[s]'})-[:has_stock_data]->(sd:stock_data{`date`: date('[d]')})"
    print(gold)
    splits = find_matches(gold)
    print(splits)



from dataclasses import dataclass

@dataclass
class Config:
    model_path: str 
    vocab_size: int
    tensor_parallel_size: int 
    temperature: float = 0.0
    max_tokens: int = 64
    logprobs: int = None 

    def to_dict(self):
        return self.__dict__



qwen_7b_config = Config("/sshfs/pretrains/Qwen/Qwen-7B-Chat", 151936, 1)
qwen_14b_config = Config("/sshfs/pretrains/Qwen/Qwen-14B-Chat", 152064, 1)

chatglm2_6b_config = Config("/sshfs/pretrains/THUDM/chatglm2-6b", 65024, 1)
chatglm3_6b_config = Config("/sshfs/pretrains/THUDM/chatglm3-6b", 65024, 1)

llama2_7b_config = Config("/sshfs/pretrains/meta-llama/Llama-2-7b-chat-hf", 32000, 1)
llama2_70b_config = Config("/sshfs/pretrains/meta-llama/Llama-2-70b-chat-hf", 32000, 4)

baichuan_7b_config = Config("/sshfs/pretrains/baichuan-inc/baichuan-7b-sft", 64000, 1)

baichuan2_7b_config = Config("/sshfs/pretrains/baichuan-inc/Baichuan2-7B-Chat",  125696, 1)
baichuan2_13b_config = Config("sshfs/pretrains/baichuan-inc/Baichuan2-13B-Chat", 125696, 2)

MODEL_DICT = {
    'qwen_7b': qwen_7b_config,
    'chatglm2_6b': chatglm2_6b_config,
    'chatglm3_6b': chatglm3_6b_config,
    'llama2_7b': llama2_7b_config,
    'llama2_70b': llama2_70b_config,
    'baichuan_7b': baichuan_7b_config,
    'baichuan2_7b': baichuan2_7b_config,
    'baichuan2_13b': baichuan2_13b_config,
}


import os, sys 
current_directory = os.path.dirname(os.path.realpath(__file__))
sys.path.append(current_directory)
import re
from utils import OPERATOR, FUNCTION, find_functions, find_calculates, find_tag_attrs, find_orderby


def find_withs(text):
    rag_attrs = []
    calculates = []
    functions = []
    temps = []
    orderbys = []
    limits = []
    text = re.sub(r'\s+', ' ', text)  # 去除多个连续空格
    text = text.lower().replace('with ', '')  # 去除with字符
    text = re.sub(r'\s+as\s+\w+\s*', ' ', text)  # 去除自命名
    if ' limit ' in text:
        match = re.search(r'\blimit\s+(\d+)\s*(.*)$', text)
        if match:
            limit = match.group(1).strip()
            limits.append(limit)
            # print(limit)
            text = re.sub(r'\s*limit\s+\d+\s*.*$', '', text)
        else:
            pass
    sents = text.split(', ')
    sents = [i.strip() for i in sents]
    # print(sents)
    for sent in sents:
        # print(sent)
        if 'order by' in sent and "row_number()" not in sent:
            orderby = find_orderby(sent)
            orderbys.append(orderby)
            continue
        elif any(substring in sent for substring in [i + '(' for i in FUNCTION]):
            function = find_functions(sent)
            functions.append(function)
            continue
        elif any(substring in sent for substring in [' ' + i + '' for i in OPERATOR]):
            # print('sent', sent)
            calculate = find_calculates(sent)
            calculates.append(calculate)
            continue
        # 定义temp
        if not re.findall(r'[\w`]+\.[\w`]+', text):
            temps.append('temp')
            continue
        rag_attr = find_tag_attrs(sent)
        rag_attrs.append(rag_attr)

    splits = {
        'with_rag_attrs': sorted(rag_attrs),
        'with_calculates': sorted(calculates),
        'with_functions': sorted(functions),
        'with_temps': sorted(temps),
        'with_orderbys': sorted(orderbys)
    }
    return splits




if __name__ == '__main__':
    gold = "WITH c.chairman.sex AS temp1, WITH c.chairman.name AS temp2, avg(td.trade_data.closing_price) AS avg_closing_price, c.chairman.name - 1 AS temp3"
    gold = "WITH s.stock.name AS sname, sd.stock_data.closing_price AS SSDR ORDER BY SSDR DESC LIMIT 1"

    splits = find_withs(gold)
    print(splits)







# Generated from ./Sparql.g4 by ANTLR 4.9.3
from antlr4 import *
if __name__ is not None and "." in __name__:
    from .SparqlParser import SparqlParser
else:
    from SparqlParser import SparqlParser

# This class defines a complete generic visitor for a parse tree produced by SparqlParser.

class SparqlVisitor(ParseTreeVisitor):

    # Visit a parse tree produced by SparqlParser#query.
    def visitQuery(self, ctx:SparqlParser.QueryContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SparqlParser#prologue.
    def visitPrologue(self, ctx:SparqlParser.PrologueContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SparqlParser#baseDecl.
    def visitBaseDecl(self, ctx:SparqlParser.BaseDeclContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SparqlParser#prefixDecl.
    def visitPrefixDecl(self, ctx:SparqlParser.PrefixDeclContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SparqlParser#selectQuery.
    def visitSelectQuery(self, ctx:SparqlParser.SelectQueryContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SparqlParser#subSelect.
    def visitSubSelect(self, ctx:SparqlParser.SubSelectContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SparqlParser#selectClause.
    def visitSelectClause(self, ctx:SparqlParser.SelectClauseContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SparqlParser#selectModifier.
    def visitSelectModifier(self, ctx:SparqlParser.SelectModifierContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SparqlParser#selectVariables.
    def visitSelectVariables(self, ctx:SparqlParser.SelectVariablesContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SparqlParser#constructQuery.
    def visitConstructQuery(self, ctx:SparqlParser.ConstructQueryContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SparqlParser#describeQuery.
    def visitDescribeQuery(self, ctx:SparqlParser.DescribeQueryContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SparqlParser#askQuery.
    def visitAskQuery(self, ctx:SparqlParser.AskQueryContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SparqlParser#datasetClause.
    def visitDatasetClause(self, ctx:SparqlParser.DatasetClauseContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SparqlParser#whereClause.
    def visitWhereClause(self, ctx:SparqlParser.WhereClauseContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SparqlParser#solutionModifier.
    def visitSolutionModifier(self, ctx:SparqlParser.SolutionModifierContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SparqlParser#groupClause.
    def visitGroupClause(self, ctx:SparqlParser.GroupClauseContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SparqlParser#groupCondition.
    def visitGroupCondition(self, ctx:SparqlParser.GroupConditionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SparqlParser#havingClause.
    def visitHavingClause(self, ctx:SparqlParser.HavingClauseContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SparqlParser#havingCondition.
    def visitHavingCondition(self, ctx:SparqlParser.HavingConditionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SparqlParser#orderClause.
    def visitOrderClause(self, ctx:SparqlParser.OrderClauseContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SparqlParser#orderCondition.
    def visitOrderCondition(self, ctx:SparqlParser.OrderConditionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SparqlParser#limitOffsetClauses.
    def visitLimitOffsetClauses(self, ctx:SparqlParser.LimitOffsetClausesContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SparqlParser#limitClause.
    def visitLimitClause(self, ctx:SparqlParser.LimitClauseContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SparqlParser#offsetClause.
    def visitOffsetClause(self, ctx:SparqlParser.OffsetClauseContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SparqlParser#valuesClause.
    def visitValuesClause(self, ctx:SparqlParser.ValuesClauseContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SparqlParser#updateCommand.
    def visitUpdateCommand(self, ctx:SparqlParser.UpdateCommandContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SparqlParser#update.
    def visitUpdate(self, ctx:SparqlParser.UpdateContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SparqlParser#load.
    def visitLoad(self, ctx:SparqlParser.LoadContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SparqlParser#clear.
    def visitClear(self, ctx:SparqlParser.ClearContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SparqlParser#drop.
    def visitDrop(self, ctx:SparqlParser.DropContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SparqlParser#create.
    def visitCreate(self, ctx:SparqlParser.CreateContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SparqlParser#add.
    def visitAdd(self, ctx:SparqlParser.AddContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SparqlParser#move.
    def visitMove(self, ctx:SparqlParser.MoveContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SparqlParser#copy.
    def visitCopy(self, ctx:SparqlParser.CopyContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SparqlParser#insertData.
    def visitInsertData(self, ctx:SparqlParser.InsertDataContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SparqlParser#deleteData.
    def visitDeleteData(self, ctx:SparqlParser.DeleteDataContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SparqlParser#deleteWhere.
    def visitDeleteWhere(self, ctx:SparqlParser.DeleteWhereContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SparqlParser#modify.
    def visitModify(self, ctx:SparqlParser.ModifyContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SparqlParser#deleteClause.
    def visitDeleteClause(self, ctx:SparqlParser.DeleteClauseContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SparqlParser#insertClause.
    def visitInsertClause(self, ctx:SparqlParser.InsertClauseContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SparqlParser#usingClause.
    def visitUsingClause(self, ctx:SparqlParser.UsingClauseContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SparqlParser#graphOrDefault.
    def visitGraphOrDefault(self, ctx:SparqlParser.GraphOrDefaultContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SparqlParser#graphRef.
    def visitGraphRef(self, ctx:SparqlParser.GraphRefContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SparqlParser#graphRefAll.
    def visitGraphRefAll(self, ctx:SparqlParser.GraphRefAllContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SparqlParser#quadPattern.
    def visitQuadPattern(self, ctx:SparqlParser.QuadPatternContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SparqlParser#quadData.
    def visitQuadData(self, ctx:SparqlParser.QuadDataContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SparqlParser#quads.
    def visitQuads(self, ctx:SparqlParser.QuadsContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SparqlParser#quadsDetails.
    def visitQuadsDetails(self, ctx:SparqlParser.QuadsDetailsContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SparqlParser#quadsNotTriples.
    def visitQuadsNotTriples(self, ctx:SparqlParser.QuadsNotTriplesContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SparqlParser#triplesTemplate.
    def visitTriplesTemplate(self, ctx:SparqlParser.TriplesTemplateContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SparqlParser#groupGraphPattern.
    def visitGroupGraphPattern(self, ctx:SparqlParser.GroupGraphPatternContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SparqlParser#groupGraphPatternSub.
    def visitGroupGraphPatternSub(self, ctx:SparqlParser.GroupGraphPatternSubContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SparqlParser#groupGraphPatternSubList.
    def visitGroupGraphPatternSubList(self, ctx:SparqlParser.GroupGraphPatternSubListContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SparqlParser#triplesBlock.
    def visitTriplesBlock(self, ctx:SparqlParser.TriplesBlockContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SparqlParser#graphPatternNotTriples.
    def visitGraphPatternNotTriples(self, ctx:SparqlParser.GraphPatternNotTriplesContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SparqlParser#optionalGraphPattern.
    def visitOptionalGraphPattern(self, ctx:SparqlParser.OptionalGraphPatternContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SparqlParser#graphGraphPattern.
    def visitGraphGraphPattern(self, ctx:SparqlParser.GraphGraphPatternContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SparqlParser#serviceGraphPattern.
    def visitServiceGraphPattern(self, ctx:SparqlParser.ServiceGraphPatternContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SparqlParser#bind.
    def visitBind(self, ctx:SparqlParser.BindContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SparqlParser#inlineData.
    def visitInlineData(self, ctx:SparqlParser.InlineDataContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SparqlParser#dataBlock.
    def visitDataBlock(self, ctx:SparqlParser.DataBlockContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SparqlParser#inlineDataOneVar.
    def visitInlineDataOneVar(self, ctx:SparqlParser.InlineDataOneVarContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SparqlParser#inlineDataFull.
    def visitInlineDataFull(self, ctx:SparqlParser.InlineDataFullContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SparqlParser#dataBlockValues.
    def visitDataBlockValues(self, ctx:SparqlParser.DataBlockValuesContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SparqlParser#dataBlockValue.
    def visitDataBlockValue(self, ctx:SparqlParser.DataBlockValueContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SparqlParser#minusGraphPattern.
    def visitMinusGraphPattern(self, ctx:SparqlParser.MinusGraphPatternContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SparqlParser#groupOrUnionGraphPattern.
    def visitGroupOrUnionGraphPattern(self, ctx:SparqlParser.GroupOrUnionGraphPatternContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SparqlParser#filter.
    def visitFilter(self, ctx:SparqlParser.FilterContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SparqlParser#constraint.
    def visitConstraint(self, ctx:SparqlParser.ConstraintContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SparqlParser#functionCall.
    def visitFunctionCall(self, ctx:SparqlParser.FunctionCallContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SparqlParser#argList.
    def visitArgList(self, ctx:SparqlParser.ArgListContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SparqlParser#expressionList.
    def visitExpressionList(self, ctx:SparqlParser.ExpressionListContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SparqlParser#constructTemplate.
    def visitConstructTemplate(self, ctx:SparqlParser.ConstructTemplateContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SparqlParser#constructTriples.
    def visitConstructTriples(self, ctx:SparqlParser.ConstructTriplesContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SparqlParser#triplesSameSubject.
    def visitTriplesSameSubject(self, ctx:SparqlParser.TriplesSameSubjectContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SparqlParser#propertyList.
    def visitPropertyList(self, ctx:SparqlParser.PropertyListContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SparqlParser#propertyListNotEmpty.
    def visitPropertyListNotEmpty(self, ctx:SparqlParser.PropertyListNotEmptyContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SparqlParser#verb.
    def visitVerb(self, ctx:SparqlParser.VerbContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SparqlParser#objectList.
    def visitObjectList(self, ctx:SparqlParser.ObjectListContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SparqlParser#object.
    def visitObject(self, ctx:SparqlParser.ObjectContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SparqlParser#triplesSameSubjectPath.
    def visitTriplesSameSubjectPath(self, ctx:SparqlParser.TriplesSameSubjectPathContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SparqlParser#propertyListPath.
    def visitPropertyListPath(self, ctx:SparqlParser.PropertyListPathContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SparqlParser#propertyListPathNotEmpty.
    def visitPropertyListPathNotEmpty(self, ctx:SparqlParser.PropertyListPathNotEmptyContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SparqlParser#propertyListPathNotEmptyList.
    def visitPropertyListPathNotEmptyList(self, ctx:SparqlParser.PropertyListPathNotEmptyListContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SparqlParser#verbPath.
    def visitVerbPath(self, ctx:SparqlParser.VerbPathContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SparqlParser#verbSimple.
    def visitVerbSimple(self, ctx:SparqlParser.VerbSimpleContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SparqlParser#objectListPath.
    def visitObjectListPath(self, ctx:SparqlParser.ObjectListPathContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SparqlParser#objectPath.
    def visitObjectPath(self, ctx:SparqlParser.ObjectPathContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SparqlParser#path.
    def visitPath(self, ctx:SparqlParser.PathContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SparqlParser#pathAlternative.
    def visitPathAlternative(self, ctx:SparqlParser.PathAlternativeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SparqlParser#pathSequence.
    def visitPathSequence(self, ctx:SparqlParser.PathSequenceContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SparqlParser#pathElt.
    def visitPathElt(self, ctx:SparqlParser.PathEltContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SparqlParser#pathEltOrInverse.
    def visitPathEltOrInverse(self, ctx:SparqlParser.PathEltOrInverseContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SparqlParser#pathMod.
    def visitPathMod(self, ctx:SparqlParser.PathModContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SparqlParser#pathPrimary.
    def visitPathPrimary(self, ctx:SparqlParser.PathPrimaryContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SparqlParser#pathNegatedPropertySet.
    def visitPathNegatedPropertySet(self, ctx:SparqlParser.PathNegatedPropertySetContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SparqlParser#pathOneInPropertySet.
    def visitPathOneInPropertySet(self, ctx:SparqlParser.PathOneInPropertySetContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SparqlParser#integer.
    def visitInteger(self, ctx:SparqlParser.IntegerContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SparqlParser#triplesNode.
    def visitTriplesNode(self, ctx:SparqlParser.TriplesNodeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SparqlParser#blankNodePropertyList.
    def visitBlankNodePropertyList(self, ctx:SparqlParser.BlankNodePropertyListContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SparqlParser#triplesNodePath.
    def visitTriplesNodePath(self, ctx:SparqlParser.TriplesNodePathContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SparqlParser#blankNodePropertyListPath.
    def visitBlankNodePropertyListPath(self, ctx:SparqlParser.BlankNodePropertyListPathContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SparqlParser#collection.
    def visitCollection(self, ctx:SparqlParser.CollectionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SparqlParser#collectionPath.
    def visitCollectionPath(self, ctx:SparqlParser.CollectionPathContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SparqlParser#graphNode.
    def visitGraphNode(self, ctx:SparqlParser.GraphNodeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SparqlParser#graphNodePath.
    def visitGraphNodePath(self, ctx:SparqlParser.GraphNodePathContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SparqlParser#varOrTerm.
    def visitVarOrTerm(self, ctx:SparqlParser.VarOrTermContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SparqlParser#varOrIRI.
    def visitVarOrIRI(self, ctx:SparqlParser.VarOrIRIContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SparqlParser#var.
    def visitVar(self, ctx:SparqlParser.VarContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SparqlParser#graphTerm.
    def visitGraphTerm(self, ctx:SparqlParser.GraphTermContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SparqlParser#nil.
    def visitNil(self, ctx:SparqlParser.NilContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SparqlParser#unarySignedLiteralExpression.
    def visitUnarySignedLiteralExpression(self, ctx:SparqlParser.UnarySignedLiteralExpressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SparqlParser#conditionalOrExpression.
    def visitConditionalOrExpression(self, ctx:SparqlParser.ConditionalOrExpressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SparqlParser#additiveExpression.
    def visitAdditiveExpression(self, ctx:SparqlParser.AdditiveExpressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SparqlParser#unaryAdditiveExpression.
    def visitUnaryAdditiveExpression(self, ctx:SparqlParser.UnaryAdditiveExpressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SparqlParser#relationalExpression.
    def visitRelationalExpression(self, ctx:SparqlParser.RelationalExpressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SparqlParser#relationalSetExpression.
    def visitRelationalSetExpression(self, ctx:SparqlParser.RelationalSetExpressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SparqlParser#unaryMultiplicativeExpression.
    def visitUnaryMultiplicativeExpression(self, ctx:SparqlParser.UnaryMultiplicativeExpressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SparqlParser#baseExpression.
    def visitBaseExpression(self, ctx:SparqlParser.BaseExpressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SparqlParser#multiplicativeExpression.
    def visitMultiplicativeExpression(self, ctx:SparqlParser.MultiplicativeExpressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SparqlParser#conditionalAndExpression.
    def visitConditionalAndExpression(self, ctx:SparqlParser.ConditionalAndExpressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SparqlParser#unaryNegationExpression.
    def visitUnaryNegationExpression(self, ctx:SparqlParser.UnaryNegationExpressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SparqlParser#unaryLiteralExpression.
    def visitUnaryLiteralExpression(self, ctx:SparqlParser.UnaryLiteralExpressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SparqlParser#unaryExpression.
    def visitUnaryExpression(self, ctx:SparqlParser.UnaryExpressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SparqlParser#primaryExpression.
    def visitPrimaryExpression(self, ctx:SparqlParser.PrimaryExpressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SparqlParser#builtInCall.
    def visitBuiltInCall(self, ctx:SparqlParser.BuiltInCallContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SparqlParser#regexExpression.
    def visitRegexExpression(self, ctx:SparqlParser.RegexExpressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SparqlParser#subStringExpression.
    def visitSubStringExpression(self, ctx:SparqlParser.SubStringExpressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SparqlParser#strReplaceExpression.
    def visitStrReplaceExpression(self, ctx:SparqlParser.StrReplaceExpressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SparqlParser#existsFunction.
    def visitExistsFunction(self, ctx:SparqlParser.ExistsFunctionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SparqlParser#notExistsFunction.
    def visitNotExistsFunction(self, ctx:SparqlParser.NotExistsFunctionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SparqlParser#aggregate.
    def visitAggregate(self, ctx:SparqlParser.AggregateContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SparqlParser#iriRefOrFunction.
    def visitIriRefOrFunction(self, ctx:SparqlParser.IriRefOrFunctionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SparqlParser#rdfLiteral.
    def visitRdfLiteral(self, ctx:SparqlParser.RdfLiteralContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SparqlParser#numericLiteral.
    def visitNumericLiteral(self, ctx:SparqlParser.NumericLiteralContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SparqlParser#numericLiteralUnsigned.
    def visitNumericLiteralUnsigned(self, ctx:SparqlParser.NumericLiteralUnsignedContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SparqlParser#numericLiteralPositive.
    def visitNumericLiteralPositive(self, ctx:SparqlParser.NumericLiteralPositiveContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SparqlParser#numericLiteralNegative.
    def visitNumericLiteralNegative(self, ctx:SparqlParser.NumericLiteralNegativeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SparqlParser#booleanLiteral.
    def visitBooleanLiteral(self, ctx:SparqlParser.BooleanLiteralContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SparqlParser#string.
    def visitString(self, ctx:SparqlParser.StringContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SparqlParser#iri.
    def visitIri(self, ctx:SparqlParser.IriContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SparqlParser#prefixedName.
    def visitPrefixedName(self, ctx:SparqlParser.PrefixedNameContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SparqlParser#blankNode.
    def visitBlankNode(self, ctx:SparqlParser.BlankNodeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SparqlParser#anon.
    def visitAnon(self, ctx:SparqlParser.AnonContext):
        return self.visitChildren(ctx)



del SparqlParser


import torch
import numpy as np

def safesoftmax(x, dim=-1):
    """ equals `torch.softmax(x, dim=-1)` """
    if type(x) is torch.Tensor:
        return torch.softmax(x-x.max(-1, keepdim=True)[0], dim)
    else:
        x = np.exp(x - np.max(x))
        return x/x.sum()

def feature_transform(feature):
    """ 变换完之后，方差1，均值0 """
    head_feature = torch.log(torch.sqrt(torch.sigmoid(feature)))
    std, mean = torch.std_mean(head_feature, 1, keepdim=True)
    return (head_feature-mean) / std


def select_span_rep(s,e,hidden):
    """
    先将hidden第一位拼接上0向量，使用的index也加1，再gather

    Args:
        s (torch.Tensor): (bsz,)
        e (torch.Tensor): (bsz,)
        hidden (torch.Tensor):  (bsz,seqlen, dim)

    Returns:
        span_rep: (bsz, seq_max+1, dim)
        mask:     (bsz, seq_max+1)          1的位置表示是Mask掉了

    Example:
        >>> s = torch.tensor([4,10,0,1])
        >>> e = torch.tensor([10,12,0,5])
        >>> hidden = torch.randn(4,30,2)
        >>> select_span_rep(s,e,hidden)

    """
    assert torch.all(e.ge(s)), f"结束索引必须大于等于开始索引: e:{e.cpu().numpy()}, s:{s.cpu().numpy()}"
    bsz,seqlen, dim = hidden.size()
    seq_max = torch.max(e-s)+1             # 最小是1最大不限>_>不能超过最大长度
    index = torch.tensor(
        [ [i+1 if i <= _e else 0  for i in range(_s, _s+seq_max)] for _s,_e in zip(s,e)],  # i+1 因为最前面要拼上一个0向量作为pad的表示
        device=hidden.device
    )[:,:,None].repeat((1,1,dim))     # index 不能超过seqlen

    mask = torch.where(index==0, torch.ones_like(index), torch.zeros_like(index))          # index等于0的位置是pad需要mask掉，设置为1
    shift_right = torch.cat([torch.zeros(bsz,1,dim, device=hidden.device), hidden], dim=1)
    span_rep = shift_right.gather(1, index)
    return span_rep, mask[:,:,0]


def select_func(m, i):
    """
        从向量序列中选取对应位置的值
    Args:
        m ( Tensor ): (bsz,seqlen,...)
        i ( Tensor ): (bsz,)

    Returns:
        torch.Tensor: (bsz, 1)
    """
    return torch.stack([ l[idx] for l, idx in zip(m, i)])


def get_max_startend(start_logits, end_logits, max_a_len=None):
    """
        MRC寻找最大概率的区间
    """
    se_sum = end_logits[:,None,:] + start_logits[:,:,None]
    # 限制值的范围是s<e, 最大长度为 max_a_len         # s!=e
    mask = torch.tril(torch.triu(torch.ones_like(se_sum), 0), max_a_len)   # (0,0) is necessary !!!!!!!!!!!!
    r = (mask * se_sum).masked_fill_((1-mask).bool(), float('-inf'))    # score值全是负的，导致0 > score，选出来s>e了
    start_max, end_max = r.max(2)[0].max(1)[1], r.max(1)[0].max(1)[1]
    assert all(end_max.ge(start_max)), "结束索引必须大于等于开始索引"
    return start_max, end_max


def label_smoothing(inputs, epsilon=0.1):
    '''Applies label smoothing. See 5.4 and https://arxiv.org/abs/1512.00567.
    inputs: 3d tensor. [N, T, V], where V is the number of vocabulary.
    epsilon: Smoothing rate.
    Label Smoothing起到的作用实际上是抑制了feature norm,
    此时网络输出的softmax prob永远无法达到设定的 1-α/(K-1) ,
    loss曲面上不再存在平缓区域,处处都有较大的梯度指向各个类中心,
    所以特征会更加聚拢. https://zhuanlan.zhihu.com/p/302843504
    For example:
    ```
    >>> y = torch.tensor([[1,0,0,0,0,0]])
    >>> label_smoothing(y)
        tensor([[0.9200, 0.0200, 0.0200, 0.0200, 0.0200, 0.0200]])
    ```
    '''
    num_classes = inputs.size(-1)
    return ((1-epsilon) * inputs) + (epsilon / (num_classes-1))


def mean_pooling(hidden_states, attention_mask):
    """输入一个向量序列， 根据attention_mask计算均值向量
        值为1的地方保留， 值为0的位置被mask
    Args:
        hidden_states (Tensor):  (bsz, seqlen, dim)
        attention_mask (Tensor): (bsz, seqlen)

    Returns:
        Tensor: (bsz, dim)
    """
    input_mask_expanded = attention_mask.unsqueeze(-1).expand(hidden_states.size()).float()
    sum_embeddings = torch.sum(hidden_states * input_mask_expanded, 1)
    sum_mask = torch.clamp(input_mask_expanded.sum(1), min=1e-9)
    return sum_embeddings / sum_mask


import torch 
import pytorch_lightning as pl
from pytorch_lightning.callbacks import Callback, ModelCheckpoint
from tqdm import tqdm
from qdls.utils import print_string

class BatchEvalCallback(Callback):
    """ 
        在训练过程中，validataion epoch 结束后，使用训练好的模型生成一些文本打印出来，用于人工评估模型效果
    """
    def on_validation_epoch_end(self, trainer: "pl.Trainer", pl_module: "pl.LightningModule") -> None:
        pl_module.model.eval()
        with torch.no_grad():
            try:
                s = trainer.datamodule.trainset[0]['input_text']
            except:
                s = 'This is the default prompt for evalation. We only calc val_loss during validation because it will be too slow to generate text for all valset.' + \
                    'So we use this default prompt to generate only one sentence for evaluation.'
            td = trainer.datamodule.tokenizer(s, return_tensors='pt').to(pl_module.model.device)
            pred_ids = pl_module.model.generate(
                input_ids=td.input_ids, attention_mask=td.attention_mask,
                max_new_tokens=384, do_sample=False, num_beams=1,
            )
            print_string("eval: pred_examples:")
            print(trainer.datamodule.tokenizer.batch_decode(pred_ids, clean_up_tokenization_spaces=True, skip_special_tokens=True)," ||| ", s)
        
        pl_module.model.train()



from weakref import proxy
class HuggingfaceModelCKPTCallback(ModelCheckpoint):
    """ 
        修改自 pytorch_lightning.callbacks.ModelCheckpoint 使得其使用 save_pretrained 保存模型权重
        适用于 huggingface transformers 模型 以及 peft 模型
        
    """
    def _save_checkpoint(self, trainer: "pl.Trainer", filepath: str) -> None:
        
        # trainer.save_checkpoint(filepath, self.save_weights_only)
        trainer.model.model.save_pretrained(filepath)
        trainer.datamodule.tokenizer.save_pretrained(filepath)

        self._last_global_step_saved = trainer.global_step

        # notify loggers
        if trainer.is_global_zero:
            for logger in trainer.loggers:
                logger.after_save_checkpoint(proxy(self))


# -*- coding: utf-8 -*-
# @File    :   __init__.py
# @Time    :   2023/11/19 15:14:13
# @Author  :   Qing 
# @Email   :   aqsz2526@outlook.com
######################### docstring ########################
'''
    用于ICL 的各种retriever

'''
""" 
BM25Retriever: 用于检索文本数据的BM25检索器
VectorRetriever: 用于检索文本数据的向量检索器
    rquirements:
        faiss-gpu
        sentence_transformers

"""



from ..grammar.SparqlLexer import SparqlLexer
from ..grammar.SparqlListener import SparqlListener
from ..grammar.SparqlParser import SparqlParser

from antlr4 import InputStream,CommonTokenStream
from antlr4.error.ErrorListener import ErrorListener


class MyParserErrorListener(ErrorListener):
    def syntaxError(self, recognizer, offendingSymbol, line, column, msg, e):
        raise Exception(f"Syntax Error:{msg}\n{e}\n{column}\n{line}\n{offendingSymbol}")

class MyLexerErrorListener(ErrorListener):
    def syntaxError(self, recognizer, offendingSymbol, line, column, msg, e):
        raise Exception(f"Lexer Error:{msg}\n{e}\n{column}\n{line}\n{offendingSymbol}")

def syntax_check(sparql):
    """
        using antlr4 to pre-check if generated sparql is syntax correct
        using self-defined ErrorListener to avoid console output
    """
    input_stream = InputStream(sparql)
    lexer = SparqlLexer(input_stream)
    lexer._listeners = []
    lexer.addErrorListener(MyLexerErrorListener())
    stream = CommonTokenStream(lexer)
    parser = SparqlParser(stream)
    parser._listeners = []
    parser.addErrorListener(MyParserErrorListener())
    try:
        tree = parser.query()
    except Exception as e:
        # print(e)
        return False, e, parser
    return True, tree, parser


if __name__ == '__main__':
    
    s1 = "select {?e ?p 'asd'}"

    flag, _ = syntax_check(s1)

    print(flag)
    print("error msg:", _)




import neo4j
from neo4j.time import DateTime
from neo4j import GraphDatabase, Query
import neo4j.exceptions

from tqdm import tqdm
import datetime
from multiprocessing import Pool
from concurrent.futures import ThreadPoolExecutor

from argparse import Namespace

import signal
class TimeoutError(Exception):
    pass

# 如果内置的timeout不好用，可以使用这个
def timeout(seconds):
    def decorator(func):
        def handler(signum, frame):
            raise TimeoutError("Function call timed out")

        def wrapper(*args, **kwargs):
            signal.signal(signal.SIGALRM, handler)
            signal.alarm(seconds)

            try:
                result = func(*args, **kwargs)
            finally:
                signal.alarm(0)  # 重置信号

            return result

        return wrapper

    return decorator    

def query_neo4j(query, driver, timeout=10):
    with driver.session(database='neo4j') as session:
        q = Query(query, timeout=timeout)
        res = session.run(q).data()
        return res 

def batch_query(queries, uri, user, passwd, nthreads=8, timeout=10):
    """并行执行查询，需要传入neo4j参数，用于给每个线程创建连接

    Args:
        queries: a list of query strings
        uri: _description_
        user: _description_
        passwd: _description_
        nthreads: _description_. Defaults to 8.
        timeout: _description_. Defaults to 10.

    Returns:
        a list of results
    """
    n = len(queries)
    config = Namespace(
        neo4j_uri=uri, neo4j_user=user, neo4j_passwd=passwd, timeout=timeout
    )
    if n == 1:
        return [exec_single_query(queries[0], config)]
    else:
        return threads_execution(queries, config, nthreads if n > nthreads else n)


def exec_single_query(query, config):
    """ 
        neo4j.time.DateTime 无法序列化，因此转换为python的datetime.datetime
        
        neo4j 的 driver 是线程安全的，通过传入的config参数在每个线程创建一个driver连接
            ref: https://community.neo4j.com/t5/drivers-stacks/python-generator-already-executing/m-p/40421

        return:
            Exception 或 list
    """
    driver = GraphDatabase.driver(uri=config.neo4j_uri, auth=(config.neo4j_user, config.neo4j_passwd))
    with driver.session(database="neo4j") as session:
        q = Query(query, timeout=config.timeout)
        res = session.run(q).data()
        return res

def threads_execution(queries, nthreads=16, config=None):
    """
    Args:
        queries: a list of query strings
        config: 连接参数

    Returns:
        执行结果
    """
    if config is None:
        raise Exception(f'config is None!!! expected: argparse.Namespace(neo4j_uri="neo4j://ip:port", neo4j_user="neo4j", neo4j_passwd="xxx", timeout=10)')

    
    thread_pool = ThreadPoolExecutor(nthreads) 
    R = []
    result_iter = thread_pool.map(exec_single_query, queries, [config]*len(queries))
    for res in tqdm(result_iter, total=len(queries)):
        R.append(res)
    assert len(R) == len(queries), f"{len(R)} != {len(queries)}"
    return R 


def process_execute(queries, nproc=8, config=None):
    """多进程执行查询，速度比多线程快(3000,32测试快3倍)

    Args:
        queries: List, 要处理的数据,
        nproc: 进程数目. Defaults to 8.
        config: Neo4j连接配置. Defaults to None.

    Raises:
        Exception: config 

    Returns:
        List, 执行结果
    """
    if config is None:
        raise Exception(f'config is None!!! expected: argparse.Namespace(neo4j_uri="neo4j://ip:port", neo4j_user="neo4j", neo4j_passwd="xxx", timeout=10)')

    Results = [] 
    with Pool(nproc) as pool:
        R = {}
        for sample in queries:
            future = pool.apply_async(exec_single_query, (sample, config))
            R[future] = sample
        
        for future in tqdm(R):
            res = future.get()
            Results.append(res)
    assert len(Results) == len(queries), f"{len(R)} != {len(queries)}"
    return Results

    
def neo4j_data_serializable(obj):
    if type(obj) is DateTime:
        dt = obj
        return datetime.datetime(dt.year, dt.month, dt.day, dt.hour, dt.minute,
                        int(dt.second), int(dt.second * 1000000 % 1000000))
    elif type(obj) is neo4j.time.Date:
        return obj.isoformat()
    elif type(obj) is dict:
        return { k : neo4j_data_serializable(v) for k,v in obj.items()}
    elif type(obj) is list:
        return [ neo4j_data_serializable(_) for _ in obj]
    else:
        return obj  
    



import os 
import sys 
os.environ["TOKENIZERS_PARALLELISM"] = "true"

FILE_DIR = os.path.dirname(os.path.abspath(__file__))
PROJ_DIR = FILE_DIR[:FILE_DIR.index('src')]
sys.path.append(PROJ_DIR)

import torch 
torch.set_float32_matmul_precision('high')
import pytorch_lightning as pl
pl.seed_everything(3407)

import fire
from omegaconf import OmegaConf 
from qdls.utils import print_config, print_string
from qdls.reg.register import import_all_modules_for_register
import_all_modules_for_register()

from frame import train_model, predict_model, data_dev

def main(config_file=None, mode=None, version='default_version', **kwargs):
    if config_file is None:
        raise Exception(f"must specify a configuration file to start!")
    
    config = OmegaConf.load(config_file)
    config.version = version 
    if mode is not None:
        config.mode = mode 
    
    cli_str = [ f"{k}={v}" for k,v in kwargs.items() ]
    config = OmegaConf.unsafe_merge(config, OmegaConf.from_cli(cli_str))
    print_config(config)
 
    if config.mode == 'train':
        train_model(config)
    elif config.mode in ['predict']:
        predict_model(config)
    elif config.mode == 'data_dev': # for dataloader development
        data_dev(config)
    else:
        raise Exception(f"mode `{config.mode}` is not implemented yet!")


if __name__ == '__main__':
    
    fire.Fire(main)


# from qdls.reg.register import registers
from ..register import registers


class Model:
    pass 

@registers.model.register("model1")
class Model1(Model):
    pass

@registers.model.register("model2")
class Model2(Model):
    pass



import re
import sys
import os
import json

current_directory = os.path.dirname(os.path.realpath(__file__))
sys.path.append(current_directory)
from match_evaluate import find_matches
from where_evaluate import find_wheres
from with_evaluate import find_withs
from return_evaluate import find_returns
import pandas as pd
from pprint import pprint
from utils import special_sort


def text_parts(text):
    match_pattern = re.compile(r'match (.*?)(?=where|return|with|$)')
    where_pattern = re.compile(r'where (.*?)(?=with|return|$)')
    with_pattern = re.compile(r'with (.*?)(?=match|return|$)')
    return_pattern = re.compile(r'return (.*)$')

    match_part = [i.strip() for i in match_pattern.findall(text)]
    where_part = [i.strip() for i in where_pattern.findall(text)]
    with_part = [i.strip() for i in with_pattern.findall(text)]
    return_part = [i.strip() for i in return_pattern.findall(text)]

    parts = {
        'match': match_part,
        'where': where_part,
        'with': with_part,
        'return': return_part,
    }
    return parts


def combine_match2where(split):
    match_where = split['match']['match_where']
    split['where']['where_calculates'].extend(match_where)
    del split['match']['match_where']
    return split


def post_split(split):
    # sort+set+去空格
    for key, value in split.items():
        for k, v in value.items():
            split[key][k] = special_sort(set([vv.replace(' ', '') for vv in v if type(vv) is str]))
    return split


def combine_split(split):
    # 合并
    for key, value in split.items():
        if value:
            merged_dict = value[0]
            for d in value:
                for k, v in d.items():
                    merged_dict[k].extend(v)
            split[key] = merged_dict
    return split


def split(text):
    # print(text)
    text = text.lower()  # 小写化处理
    text = text.replace('\n', ' ')  # 去除换行
    text = text.replace('"', "'")  # 统一引号
    text = re.sub(r'\s+', ' ', text)  # 去除多个连续空格
    text = text.replace("( ", "(").replace(" )", ")").replace(". ", ".")  # 去除多余空格
    parts = text_parts(text)
    # print(parts)

    match_splits = [{'match_tags': [],'match_where': []}]
    where_splits = [{'where_calculates': [], 'where_temps': []}]
    with_splits = [{'with_rag_attrs': [], 'with_calculates': [], 'with_functions': [], 'with_temps': [], 'with_orderbys': []}]
    return_splits = [{'return_rag_attrs': [], 'return_calculates': [], 'return_functions': [], 'return_orderby': [], 'return_temps': [], 'return_limit': []}]

    for part in parts['match']:
        match_split = find_matches(part)
        match_splits.append(match_split)
        # print(1)
    for part in parts['where']:
        where_split = find_wheres(part)
        where_splits.append(where_split)
        # print(2)
    for part in parts['with']:
        with_split = find_withs(part)
        with_splits.append(with_split)
        # print(3)
    for part in parts['return']:
        return_split = find_returns(part)
        return_splits.append(return_split)
        # print(4)
    split = {
        'match': match_splits,
        'where': where_splits,
        'with': with_splits,
        'return': return_splits,
    }
    # pprint(split)
    split = combine_split(split)
    # pprint(split)
    split = combine_match2where(split)
    # pprint(split)
    split = post_split(split)
    return split


def eval_in(list1, list2):
    list1 = [i for i in list1 if 'date' not in i]
    list2 = [i for i in list2 if 'date' not in i]
    return all(elem in list2 for elem in list1)


def del_section(dict, key_list):
    for key in key_list:
        if key in dict:
            del dict[key]
    return dict


def evaluate_em(gold, pre, escape_part=[], escape_section=[]):
    '''
    EM-evaluation for cypher
    :param gold: cypher1
    :param pre: cypher2
    :param escape_part: string-list escape_part = ["match", "where", "with", "return"]
    :param escape_section: string-list
    escape_section = ["match_tags",
    "where_calculates", "where_temps",
    "with_rag_attrs", "with_calculates", "with_functions", "with_temps",
    "return_rag_attrs", "return_calculates", "return_functions", "return_orderby", "return_temps", "return_limit"]
    :return: evaluation_result, gold_split_dict, pre_split_dict, gold_split_dict_error, pre_split_dict_error
    '''
    splits_gold = split(gold)
    splits_pre = split(pre)

    for key in splits_gold.keys():
        if key in escape_part:
            continue
        del_section(splits_gold[key], escape_section)
        del_section(splits_pre[key], escape_section)
        if splits_gold[key] != splits_pre[key]:
            return False, splits_gold, splits_pre, splits_gold[key], splits_pre[key]
    return True, splits_gold, splits_pre, None, None


def read_json_data(source_file):
    with open(source_file, "r", encoding="utf8") as f:
        dataset = json.load(f)
    return dataset


if __name__ == '__main__':

    cypher = ("MATCH (s:stock{name: '[s]'})-[:has_stock_data]->(sd:stock_data{`date`: date('[d]')}) "
              "WITH sd.stock_data.volume AS temp "
              "MATCH (s:stock)-[:has_stock_data]->(sd:stock_data{`date`: date('[d]')}) "
              "WHERE sd.stock_data.volume == temp "
              "RETURN s.stock.name LIMIT 10")
    cypher1 = ("MATCH (s:stock{name: '[s]'})-[:has_stock_data]->(sd:stock_data{`date`: date('[d]')}) "
              "WITH sd.stock_data.volume AS temp "
              "MATCH (s:stock)-[:has_stock_data]->(sd:stock_data{`date`: date('[d]')}) "
              "WHERE sd.stock_data.volume == temp "
              "RETURN s.stock.name LIMIT 100")
    cypher2 = ("MATCH (pof1:public_offering_fund{name: '[f1]'})<-[:manage]-(fm:fund_manager)-[:manage]->(pof2:public_offering_fund{name: '[f2]'}) "
              "RETURN fm.fund_manager.name")


    splits_cypher = split(cypher)
    pprint(splits_cypher)

    splits_cypher = split(cypher1)
    pprint(splits_cypher)

    splits_cypher = split(cypher2)
    pprint(splits_cypher)

    result, _, _, _, _ = evaluate_em(cypher, cypher1, escape_section=['return_limit'])
    print(result)
    result, _, _, _, _ = evaluate_em(cypher, cypher2)
    print(result)


import os,sys 
FILE_DIR = os.path.dirname(os.path.abspath(__file__))
PROJ_DIR = FILE_DIR[:FILE_DIR.index('src')]
sys.path.append(PROJ_DIR)

from tqdm import tqdm
import pytorch_lightning as pl
from pytorch_lightning.loggers import TensorBoardLogger
from pytorch_lightning.callbacks import ModelCheckpoint, EarlyStopping

from qdls.data import load_json, save_json
from qdls.utils import print_dict
from qdls.reg.register import registers
from qdls.reg.datamodules.base_dm import BaseDataModule


from src.models import *
from src.pl_callbacks import *
from src.data_utils import *


def train_model(config):
    dm = BaseDataModule(config)
    dm.prepare_data()
    dm.setup("fit")
    print_dict(dm.trainset[0])
    # dm.valset = dm.valset.shard(index=0, num_shards=10)
    # print(f"len dm.valset: {len(dm.valset)}")

    model = registers.model.get(config.model_type)(config)

    logger = TensorBoardLogger(
        save_dir="./lightning_logs/",
        name=None,                # 指定experiment, ./lightning_logs/exp_name/version_name
        version=config.version,   # 指定version, ./lightning_logs/version_name
    )
    CUR_DIR = os.getcwd()
    dirname = os.path.join(CUR_DIR, "./lightning_logs/", config.version)
    ckpt_callback = HuggingfaceModelCKPTCallback(
        dirpath=dirname,
        filename="{epoch}_{val_loss:.4f}",   # 模型保存名称， epoch信息以及验证集分数
        monitor='val_loss',
        mode='min',
        save_top_k=2,
        verbose=True,
    )
    config.total_steps = len(dm.train_dataloader()) * config.trainer.max_epochs \
        if 'max_epochs' in config.trainer else config.trainer.max_steps
    es = EarlyStopping(monitor='val_loss', mode='min', patience=3)

    trainer = pl.Trainer(
        accelerator='gpu', 
        num_nodes=config.trainer.num_nodes,                # 一台机器
        devices=config.trainer.devices,                  # 两张显卡
        strategy=config.trainer.strategy,
        # strategy='deepspeed_stage_2_offload', # does not support AdamW
        precision=config.trainer.get('precision', '32-true'),
        logger=logger,
        log_every_n_steps=10,
        callbacks=[ckpt_callback,es, BatchEvalCallback(), ],
        accumulate_grad_batches=config.trainer.accumulate_grads,
        val_check_interval=config.trainer.get('val_check_interval', 1.0),
        check_val_every_n_epoch=config.trainer.get('check_val_every_n_epoch', None),  # If you want to validate based on the total training batches,
        max_steps=config.trainer.get('max_steps', -1),
        max_epochs=config.trainer.get('max_epochs', None),
        # limit_train_batches=60,
        # limit_val_batches=5
    )

    trainer.fit(model, datamodule=dm)


def predict_model(config):
    dm = BaseDataModule(config)
    dm.prepare_data('test')
    dm.setup("test")
    print_dict(dm.testset[0])

    device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
    config.pretrained = config.pred.ckpt_path if config.pred.ckpt_path is not None else config.pretrained
    print_string(f"ckpt model {config.pretrained} set to config.pretrained")
    model = registers.model.get(config.model_type)(config).to(device) 
    T = [] 
    with torch.no_grad():
        for batch in tqdm(dm.test_dataloader()):
            batch = { k:v.to(device) if type(v) is torch.Tensor else v for k,v in batch.items()}
            pred_ids = model.generate(
                input_ids=batch['input_ids'],
                attention_mask=batch['attention_mask'],
                num_beams=1, do_sample=False,
                max_new_tokens=config.pred.get('max_new_tokens', 256),
            )
            texts = dm.tokenizer.batch_decode(pred_ids, skip_special_tokens=True, clean_up_tokenization_spaces=True)
            T.extend(texts)
    R = []
    for sample, pred in zip(dm.testset, T):
        d = {
            k:v for k,v in sample.items() if k not in ['input_ids', 'attention_mask', 'labels']
        }
        d['pred'] = pred 
        R.append(d)
    save_json(R, config.pred.result_path)


def data_dev(config):
    pass 



# -*- coding: utf-8 -*-
# @File    :   kqa_eval.py
# @Time    :   2023/06/26 16:36:09
# @Author  :   Qing 
# @Email   :   aqsz2526@outlook.com
######################### docstring ########################
'''
    elvauate KQA Cypher
'''

import os 
import neo4j
from neo4j import GraphDatabase, Query

from qdls.data import load_json, save_json
from qdls.gql.cypher.utils.syntax import syntax_check

from concurrent.futures import ThreadPoolExecutor
from multiprocessing import Pool
from argparse import Namespace
from tqdm import tqdm

from qdls.utils import print_string

import json 
from neo4j import GraphDatabase, Query
from .execute import timeout


def exec_single_query(query, config):
    """ 
        neo4j 的 driver 是线程安全的，通过传入的config参数在每个线程创建一个driver连接
            ref: https://community.neo4j.com/t5/drivers-stacks/python-generator-already-executing/m-p/40421

        return:
            返回查询的结果
            超时等 Exception 不处理 
    """

    driver = GraphDatabase.driver(uri=config.neo4j_uri, auth=(config.neo4j_user, config.neo4j_passwd))
    with driver.session(database="neo4j") as session:
        q = Query(query, timeout=config.timeout)
        res = session.run(q).data()
        return res



units = ['metre', 'pound', 'minute', 'United States dollar', 'mile', '1', 'centimetre', 'square metre', 'square kilometre', 
'hour', 'second', 'kilogram', 'inch', 'hertz', 'foot', 'Japanese yen', 'year', 'square mile', 'kilometre', 'act', 'audio track',
 'millimetre', 'Danish krone', 'hectare', 'euro', 'Argentine peso', 'beats per minute', 'Nigerian naira', 'barrels per day', 
 'kilogram per cubic metre', 'kilometre per hour', 'month', 'annum', 'Singapore dollar', 'cubic metre per second', 'Czech koruna', 
 'day', 'pound sterling', 'percentage', 'acre', 'workforce', 'square foot', 'South Korean won', 'Canadian dollar', 'Indian rupee', 
 'reservist', 'volt', 'Russian ruble', 'active duty military personnel', 'son', 'atomic mass unit', 'Hong Kong dollar', 
 'years old', 'drawing', 'book', 'gram per cubic centimetre', 'painting', 'gram per kilogram', 'sculpture', 'degree Celsius', 
 'cubic hectometre', 'photograph', 'yottagram', 'gram per mole', 'geographic map', 'astronomical unit', 'Australian dollar', 
 'degree', 'Philippine peso', 'chapter', 'Deutsche Mark', 'cubic kilometre', 'gigawatt', 'solar radius', 'kelvin', 
 'disability-adjusted life year', 'solar mass', 'degree Fahrenheit', 'millimeter of mercury', 'milligram per cubic meter', 
 'volume percent', 'electronvolt', 'metre per second', 'hectopascal', 'joule per mole-kelvin', 'sol', 'daughter', 
 'kilojoule per mole', 'watt per meter kelvin', 'gigapascal', 'cubic metre per ton', 'picometre', 'amperes per volt meter', 
 'millipascal-second', 'Caribbean guilder', 'megajoule per kilogram', 'liter per kilogram', 'grams per cubic meter', 
 'Lebanese pound', 'gram per 100 gram of solvent', 'square versta', 'gram', 'tonne', 'week', 'joule per tesla', 'country', 
 'Geary–Khamis dollar', 'cases per 100000 person-years', 'European Currency Unit', 'Swiss franc', 'hryvnia', 'Swedish krona', 
 'crore', 'Thousands', 'gigabyte', 'torr', 'movie theater', 'academic term']

def serialize_result(res):
    """现在的方法比较日期有BUG，使用此函数转换后再比较  """
    if type(res) is neo4j.time.Date:
        return res.isoformat()
    elif type(res) is dict:
        return {k:serialize_result(v) for k,v in res.items()}
    elif type(res) is list:
        return [serialize_result(_) for _ in res]
    else:
        return str(res)

def compare_answer(res, ans, query):
    
    unit  = None  
    if ans is not None:
        if query.strip()[-7:] == "type(p)" or query.strip()[-10:] == "type ( p )": #问的是关系
            if ans == "position played on team / speciality":
                ans = '_speciality'
            if " " in ans:
                ans = ans.replace(" ", "_")
            if "-" in ans or "(" in ans or "\xa0" in ans or "." in ans or "'" in ans:
                ans = f"`{ans}`"
            elif any(u in ans for u in units) and " " in ans:  # 答案中包含单位, fix_bug:有单位则须有空格
                if "1" in ans and ans.replace(" ", "").isnumeric():  # ISBN 等如 '0000 0000 4361 9306'
                    pass 
                else:
                    try:
                        n, *unit_parts = ans.split(" ")
                        unit = " ".join(unit_parts)
                        if unit in units:
                            ans = f'{float(n)}'
                    except:
                        # print(f"execution answer units converting failed: {ans}, {[ u for u in units if u in ans]}")
                        pass 
    result_str = str(serialize_result(res))
    if len(res) == 0:
        matched = True if ans == 'no' else False # 查询不到结果 答案是 no 时视为匹配
    elif ans is None:
        matched = False
    else:
        if ans in result_str:
            if unit is not None and unit not in result_str:
                matched = False
            else:
                matched = True
        elif ans == 'no' and ('False' in result_str or "None" in result_str):  # 值节点不存在该属性名，则对应null
            matched = True 
        elif ans == 'yes' and 'True' in result_str:
            matched = True 
        else:
            matched = False
    
    return matched 

def exec_one_sample_cypher(sample, config, key='pred'):
    """ 不进行 try except 处理，直接抛出异常 """
    if key not in sample:
        return False, "no_pred", []
    query = sample[key]
    res = []
    if not syntax_check(query)[0]:
        is_correct = False
        exec_info = 'syntax_error'
    else:        
        res = exec_single_query(query, config)
        is_correct = compare_answer(res, sample.get('answer', None), query)
        exec_info = 'match' if is_correct else 'mismatch'

    return is_correct, exec_info, json.dumps(res, ensure_ascii=False, default=str)





def process_execute(queries, nproc=8, config=None, key='cypher'):
    
    assert config is not None, f"config is None, please set config first"

    Results = [] 
    if nproc == 1:
        print_string("sequential execution")
        for sample in tqdm(queries):
            Results.append(exec_one_sample_cypher(sample, config, key))
    else:
        with Pool(nproc) as pool:
            R = {}
            for sample in queries:
                future = pool.apply_async(exec_one_sample_cypher, (sample, config, key))
                R[future] = sample
            
            for future in tqdm(R):
                try:
                    res = future.get()
                except Exception as e:
                    print(e)
                    print(R[future])
                    print("+"*80)
                    res = False, f"Neo4jError: {R[future]['pred']}"
                Results.append(res)
    assert len(Results) == len(queries), f"{len(R)} != {len(queries)}"
    return Results


def cypher_exec_eval(path, nproc=16, key='pred', resave=False, post_process_fn=None, config=None):
    """ 
        path: 可以是字符串（读取文件）也可以是读取后的对象list 
        key: 每一个sample字典中, 预测的sparql语句的key
        nproc: 并行处理的进程数
        resave: bool or string  是否重新保存结果, 保存在原目录, 文件名前加上exec_
        post_process_fn: 对每一个sample进行后处理的函数，返回一个新的sample
    """
    data = load_json(path) if type(path) is str else path 
    # queries = [ s['generated'] for s in data]
    # Results = threads_execution(data, nthreads=32)
    if post_process_fn is None:
        print_string("post_process_fn is None, make sure `key` do not need to be processed")
    if post_process_fn is not None:
        data = [ post_process_fn(s) for s in data]
    Results = process_execute(data, nproc=nproc, config=config, key=key)
    
    is_right = [ 1 for match, res_str in Results if match is True ]
    acc = 100 * len(is_right)/ len(Results)
    print(f"{len(is_right)} of {len(Results)} is correct, {acc:.3f}")

    if resave:
        if type(resave) is not str:
            assert type(path) is str, "resave=True, path should be a string, or set resave=filepath"
        else:
            path = resave

        for sample, exec_res in zip(data, Results):
            sample['exec'] = exec_res
        print_string(f"resaving to {resave}")
        new_filename = "exec_" + os.path.basename(resave)
        new_path = os.path.join(os.path.dirname(resave), new_filename)
        save_json(data, new_path)

    return acc, Results



if __name__ == '__main__':

    config = Namespace(neo4j_uri="neo4j://127.0.0.1:28892", neo4j_user="neo4j", neo4j_passwd="kqa", timeout=10)
    pass 




from pygments.lexers import get_lexer_by_name

from pygments.token import *
from antlr4.tree.Tree import TerminalNodeImpl
from antlr4 import InputStream, CommonTokenStream

from ..grammar.CypherParser import CypherParser
from ..grammar.CypherLexer import CypherLexer
from ...utils import parse_layer

def parse_cypher(cypher, tree_only=False):
    """ 返回输入 Cypher 语句的AST树 
        AST 树、序列化的树
        便利后得到的 每个token 及其对应的父节点
    """
    input_stream = InputStream(cypher)
    # print(input_stream)
    lexer = CypherLexer(input_stream)
    stream = CommonTokenStream(lexer)
    parser = CypherParser(stream)
    tree = parser.oC_Cypher()
    tree.toStringTree()
    if tree_only:
        return tree, parser  
    parts, parents = [], []
    # print(tree)
    # traverse(tree, parser.ruleNames)
    parse_layer(tree, parser.ruleNames, parts, parents)
    # print("after parsing", tree)
    # for token, par in zip(parts, parents):
        # print(token, "\t", par)
    s = tree.toStringTree(recog=parser)
    return tree, s, parts, parents 



def get_cypher_nodes(tree, parser):
    """遍历语法树 返回树字符串、节点列表、  

    Args:
        tree: 
        parser: 
    """
    parts, parents = [], []
    parse_layer(tree, parser.ruleNames, parts, parents)
    s = tree.toStringTree(recog=parser)
    return s, parts, parents 

def split_cypher(query, lexer=None):
    """使用lexer将查询语句分词（语义单元级别

    Args:
        query: 查询字符串，可以包含语法错误
        lexer: pygments的lexer. Defaults to cy_lexer.

    Returns:
        a list of strings 
    """
    if lexer is None:
        try:
            lexer = get_lexer_by_name('py2neo.cypher')
        except:
            raise Exception(f"lexer is not set! Using py2neo failed, run `pip install py2neo`")

    pred_units = []
    for tag, substr in lexer.get_tokens(query):
        if tag is Token.Error:
            # return -1
            pred_units.append("错")
        # print(tag, substr)
        if substr.strip() != "":
            pred_units.append(substr)

    return pred_units

